import os
import subprocess
import time
from typing import Optional
import numpy as np
from functools import wraps

def timer(c=None):
    """
    Decorator to show the time spent on a function.

    Args:
        c (Config, optional): config object

    Once decorated, only processor with ID :code:`c.pid_show` in :code:`c.comm` will run the timer in the function.
    """
    def decorator(func):
        if c is not None and not getattr(c, 'timer', True):
            # if the config states timer=False, just return original func
            return func

        @wraps(func)
        def wrapper(*args, **kwargs):
            t0 = time.time()
            try:
                return func(*args, **kwargs)
            finally:
                t1 = time.time()
                if c is None or (hasattr(c, 'comm') and c.comm.Get_rank() == getattr(c, 'pid_show', 0)):
                    print(f"timer: {func.__name__} took {t1 - t0} seconds", flush=True)
        return wrapper

    return decorator

def progress_bar(task_id: int, ntask: int, width: int=50):
    """
    Generate a progress bar based on task_id and ntask.

    Args:
        task_id (int): Current task index, from 0 to ntask-1
        ntask (int): Total number of tasks
        width (int): The length of the progress bar (number of characters)

    Returns:
        str: The progress bar msg to be shown.
            Will require the print command with end="" option so that new line updated is overwritting the old line.
    """
    if ntask==0:
        progress = 1
    else:
        progress = (task_id+1) / ntask

    ##the progress bar looks like this: .....  | ??%
    pstr = '\r{:{}}| '.format('.'*int(np.ceil(progress * width)), width)

    ##add the percentage completed at the end
    pstr += '{:.0f}%'.format(100*progress)

    return pstr

def print_with_cache(msg):
    ##previous message is cached so that new message is displayed only
    ##when it's different from the previous one (avoid redundant output)
    if not hasattr(print_with_cache, 'prev_msg'):
        print_with_cache.prev_msg = ''

    ##only show at most nmsg messages
    if msg != print_with_cache.prev_msg:
        print(msg, flush=True, end="")
        print_with_cache.prev_msg = msg

def watch_files(files, timeout=1000, check_dt=1):
    ##wait for file in files to appear, check every check_dt seconds
    ##if timeout seconds passed but still file not found, raise error
    if isinstance(files, list):
        file_list = files
    else:
        file_list = [files]
    elapsed_t = 0
    while file_list:
        file_list = [f for f in file_list if not os.path.exists(f)]
        time.sleep(check_dt)
        elapsed_t += check_dt
        if elapsed_t > timeout:
            raise RuntimeError(f"watch_files: timed out waiting for files {file_list}")

def watch_log(logfile, keyword, timeout=1000, check_dt=1):
    ##wait for keyword to appear in a logfile (indicating success in completion)
    ##check every check_dt seconds
    ##if logfile size grows (some active output is happening), reset the timer
    ##if timeout is reached, raise error
    elapsed_t = 0
    n0 = count_lines_in_file(logfile)
    while not find_keyword_in_file(logfile, keyword):
        time.sleep(check_dt)
        elapsed_t += check_dt
        n1 = count_lines_in_file(logfile)
        if n1 > n0:
            elapsed_t = 0
            n0 = n1
        if elapsed_t > timeout:
            raise RuntimeError(f"watch_log: {logfile} remain stagnant for {timeout} seconds, while waiting for keyword '{keyword}'")

def find_keyword_in_file(file, keyword):
    p = subprocess.run(f"grep '{keyword}' {file}", shell=True, capture_output=True, text=True)
    if p.stderr:
        raise RuntimeError(p.stderr)
    else:
        if p.stdout:
            return True
    return False

def count_lines_in_file(file):
    p = subprocess.run(f"wc -l {file}", shell=True, capture_output=True, text=True)
    if p.stderr:
        raise RuntimeError(p.stderr)
    else:
        n = int(p.stdout.split(' ')[0])
    return n




import numpy as np
import copy
from matplotlib.tri import Triangulation
from matplotlib import colormaps
from NEDAS.grid.grid_base import GridBase

class IrregularGrid(GridBase):
    """
    Class for handling irregular grids.

    Args:
        triangles (np.ndarray, optional):
            `None` (default), if the triangle indices `ind[grid.x.size, 3]` for the unstructured mesh are given here, the grid will skip computing triangulation.

    """
    def __init__(self, proj, x, y, bounds=None, cyclic_dim=None, distance_type='cartesian',
                 triangles=None, dst_grid=None):
        super().__init__(proj, x, y, bounds, cyclic_dim, distance_type, dst_grid)
        self.regular = False

        ##Generate triangulation, if tiangles are provided its very quick,
        ##otherwise Triangulation will generate one, but slower.
        self.x = self.x.flatten()
        self.y = self.y.flatten()
        self.npoints = self.x.size
        self.tri = Triangulation(self.x, self.y, triangles=triangles)
        self.tri.inds = np.arange(self.npoints)
        self._triangle_properties()
        dx = self._mesh_dx()
        self.dx = dx
        self.dy = dx
        self.x_elem = np.mean(self.tri.x[self.tri.triangles], axis=1)
        self.y_elem = np.mean(self.tri.y[self.tri.triangles], axis=1)
        self.Lx = self.xmax - self.xmin
        self.Ly = self.ymax - self.ymin
        if self.cyclic_dim is not None:
            self._pad_cyclic_mesh_bounds()

        self.mask = np.full(self.x.shape, False)

        self.distance_type = distance_type

        self._dst_grid = None
        if dst_grid is not None:
            self.set_destination_grid(dst_grid)

    def change_resolution_level(self, nlevel):
        raise NotImplementedError("change_resolution only works for regular grid now")

    def _mesh_dx(self):
        """
        computes averaged edge length for irregular mesh, used in self.dx, dy
        """
        t = self.tri.triangles
        s1 = np.hypot(self.x[t][:,0]-self.x[t][:,1], self.y[t][:,0]-self.y[t][:,1])
        s2 = np.hypot(self.x[t][:,0]-self.x[t][:,2], self.y[t][:,0]-self.y[t][:,2])
        s3 = np.hypot(self.x[t][:,2]-self.x[t][:,1], self.y[t][:,2]-self.y[t][:,1])
        sa = (s1 + s2 + s3)/3

        ##try to remove very elongated triangles, so that mesh dx is more accurate
        e = 0.3
        valid = np.logical_and(np.abs(s1-sa) < e*sa, np.abs(s2-sa) < e*sa, np.abs(s3-sa) < e*sa)
        if ~valid.all():
            ##all triangles are elongated, just take their mean size
            return np.mean(sa)
        else:
            ##take the mean of triangle size, excluding the elongated ones
            return np.mean(sa[valid])

    def _triangle_properties(self):
        """
        computes triangle properties for the mesh triangles
        """
        t = self.tri.triangles
        x = self.x[self.tri.inds]
        y = self.y[self.tri.inds]
        s1 = np.hypot(x[t[:,0]] - x[t[:,1]], y[t[:,0]] - y[t[:,1]])
        s2 = np.hypot(x[t[:,0]] - x[t[:,2]], y[t[:,0]] - y[t[:,2]])
        s3 = np.hypot(x[t[:,2]] - x[t[:,1]], y[t[:,2]] - y[t[:,1]])
        s = 0.5*(s1+s2+s3)
        self.tri.p = 2.0 * s  ##circumference
        self.tri.a = np.sqrt(s*(s-s1)*(s-s2)*(s-s3))  ##area
        ##circumference-to-area ratio
        ##(1: equilateral triangle, ~0: very elongated)
        self.tri.ratio =  self.tri.a / s**2 * 3**(3/2)
   
    def _pad_cyclic_mesh_bounds(self):
        ##repeat the mesh in x and y directions if cyclic, to form the wrap around geometry
        x = self.x
        y = self.y
        inds = np.arange(self.npoints)
        if 'x' in self.cyclic_dim:
            x = np.hstack([x, x-self.Lx, x+self.Lx])
            y = np.hstack([y, y, y])
            inds = np.hstack([inds, inds, inds])
        if 'y' in self.cyclic_dim:
            x = np.hstack([x, x, x])
            y = np.hstack([y, y-self.Ly, y+self.Ly])
            inds = np.hstack([inds, inds, inds])
        if 'x' in self.cyclic_dim and 'y' in self.cyclic_dim:
            x = np.hstack([x, x-self.Lx, x+self.Lx, x-self.Lx, x+self.Lx])
            y = np.hstack([y, y-self.Ly, y-self.Ly, y+self.Ly, y+self.Ly])
            inds = np.hstack([inds, inds, inds, inds, inds])
        tri = Triangulation(x, y)

        ##find the triangles that covers the domain and keep them
        triangles = []
        for i in range(tri.triangles.shape[0]):
            t = tri.triangles[i,:]
            if np.logical_and(x[t]>=self.xmin, x[t]<=self.xmax).any() and np.logical_and(y[t]>=self.ymin, y[t]<=self.ymax).any():
                triangles.append(t)
        triangles = np.array(triangles)

        ##collect the uniq grid point indices (in self.x and self.y) and assign them to the triangles
        uniq_inds = list(np.unique(triangles.reshape(-1)))
        for i in np.ndindex(triangles.shape):
            triangles[i] = uniq_inds.index(triangles[i])
        self.tri = Triangulation(x[uniq_inds], y[uniq_inds], triangles=triangles)
        self.tri.inds = inds[uniq_inds]

    def find_index(self, x_, y_):
        x_ = np.array(x_).flatten()
        y_ = np.array(y_).flatten()

        ##lon: pyproj.Proj works only for lon=-180:180
        if self.proj_name == 'longlat':
            x_ = np.mod(x_ + 180., 360.) - 180.

        ###account for cyclic dim, when points drop "outside" then wrap around
        x_, y_ = self._wrap_cyclic_xy(x_, y_)

        ##for irregular mesh, use tri_finder to find index
        tri_finder = self.tri.get_trifinder()
        triangle_map = tri_finder(x_, y_)
        inside = (triangle_map >= 0)
        indices = triangle_map[inside]

        ##internal coords are the barycentric coords (in1, in2, in3) in a triangle
        ##note: larger in1 means closer to the vertice 1!
        ##     (0,0,1) p3\
        ##            / | \
        ##           / in3. \
        ##          /  :* .   \
        ##         /in1  | in2  \
        ##(1,0,0) p1-------------p2 (0,1,0)
        vertices = self.tri.triangles[triangle_map[inside], :]

        ##transform matrix for barycentric coords computation
        a = self.tri.x[vertices[:,0]] - self.tri.x[vertices[:,2]]
        b = self.tri.x[vertices[:,1]] - self.tri.x[vertices[:,2]]
        c = self.tri.y[vertices[:,0]] - self.tri.y[vertices[:,2]]
        d = self.tri.y[vertices[:,1]] - self.tri.y[vertices[:,2]]
        det = a*d-b*c
        t_matrix = np.zeros((len(vertices), 3, 2))
        t_matrix[:,0,0] = d/det
        t_matrix[:,0,1] = -b/det
        t_matrix[:,1,0] = -c/det
        t_matrix[:,1,1] = a/det
        t_matrix[:,2,0] = self.tri.x[vertices[:,2]]
        t_matrix[:,2,1] = self.tri.y[vertices[:,2]]

        ##get barycentric coords, according to https://en.wikipedia.org/wiki/
        ##Barycentric_coordinate_system#Barycentric_coordinates_on_triangles,
        delta = np.array([x_[inside], y_[inside]]).T - t_matrix[:,2,:]
        in12 = np.einsum('njk,nk->nj', t_matrix[:,:2,:], delta)
        in_coords = np.hstack((in12, 1.-in12.sum(axis=1, keepdims=True)))

        ##index of grid nearest to (x_,y_)
        nearest = vertices[np.arange(len(in_coords), dtype=int), np.argmax(in_coords, axis=1)]

        return inside, indices, vertices, in_coords, nearest

    def _interp_weights(self, inside, vertices, in_coords):
        """
        Compute interpolation weights from the outputs of find_index
        the interp_weights are the weights (sums to 1) given to each grid vertex in self.x,y
        based on their distance to the x_,y_ points (as specified by the in_coords)
        Inputs:
        - inside, vertices, in_coords: from the output of self.find_index
        Output:
        - interp_weights: float, np.array with vertices.shape
        """
        ##use barycentric coordinates as interp weights
        interp_weights = in_coords
        return interp_weights

    def interp(self, fld, x=None, y=None, method='linear'):
        if x is None or y is None:
            ##use precalculated weights for self.dst_grid
            inside = self.interp_inside
            indices = self.interp_indices
            vertices = self.interp_vertices
            nearest = self.interp_nearest
            weights = self.interp_weights
            x = self.dst_grid.x
        else:
            ##otherwise compute the weights for the given x,y
            inside, indices, vertices, in_coords, nearest = self.find_index(x, y)
            weights = self._interp_weights(inside, vertices, in_coords)

        fld_interp = np.full(np.array(x).flatten().shape, np.nan)
        if fld.shape == self.x.shape:
            fld = fld[self.tri.inds]
            if method == 'nearest':
                # find the node of the triangle with the maximum weight
                fld_interp[inside] = fld.flatten()[nearest]
            elif method == 'linear':
                # sum over the weights for each node of triangle
                fld_interp[inside] = np.einsum('nj,nj->n', np.take(fld.flatten(), vertices), weights)
            else:
                raise NotImplementedError(f"interp method {method} is not yet available")
        elif fld.shape == self.x_elem.shape:
            fld_interp[inside] = fld[indices]
        else:
            raise ValueError("field shape does not match grid shape, or number of triangle elements")
        return fld_interp.reshape(np.array(x).shape)

    def coarsen(self, fld):
        ##find which location x_,y_ falls in in dst_grid
        if fld.shape == self.x.shape:
            inside = self.coarsen_inside
            nearest = self.coarsen_nearest
        elif fld.shape == self.x_elem.shape:
            inside = self.coarsen_inside_elem
            nearest = self.coarsen_nearest_elem
        else:
            raise ValueError("field shape does not match grid shape, or number of triangle elements")

        fld_coarse = np.zeros(self.dst_grid.x.flatten().shape)
        count = np.zeros(self.dst_grid.x.flatten().shape)
        fld_inside = fld.flatten()[inside]
        valid = ~np.isnan(fld_inside)  ##filter out nan

        ##average the fld points inside each dst_grid element
        np.add.at(fld_coarse, nearest[valid], fld_inside[valid])
        np.add.at(count, nearest[valid], 1)

        valid = (count>1)  ##do not coarse grain if only one point near by
        fld_coarse[valid] /= count[valid]
        fld_coarse[~valid] = np.nan

        return fld_coarse.reshape(self.dst_grid.x.shape)

    def plot_field(self, ax, fld,  vmin=None, vmax=None, cmap='viridis', **kwargs):
        if vmin is None:
            vmin = np.nanmin(fld)
        if vmax is None:
            vmax = np.nanmax(fld)

        if isinstance(cmap, str):
            cmap = colormaps[cmap]

        if fld.shape == self.x.shape:
            fld = fld[self.tri.inds]
        im = ax.tripcolor(self.tri, fld, vmin=vmin, vmax=vmax, cmap=cmap, **kwargs)

        self.set_xylim(ax)
        return im



import os
import numpy as np
from pyproj import Proj
from NEDAS.grid import Grid
from NEDAS.utils.conversion import dt1h
from NEDAS.utils.shell_utils import run_command, run_job, makedir
# from .namelist import namelist
# from .bin_io import read_
from NEDAS.models import Model

class WRFModel(Model):

    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        ##derived default values
        self.ref_x = self.e_we[0] / 2
        self.ref_y = self.e_sn[0] / 2

        levels = np.arange(1, self.e_vert[0]+1, 1)  ##use domain 1 setting for z levels
        level_sfc = np.array([0])
        self.variables = {
            'atmos_velocity': {'name':('u_1', 'v_1'), 'dtype':'float', 'is_vector':True, 'levels':levels, 'units':'m/s'},
            'atmos_surf_velocity': {'name':('U10', 'V10'), 'dtype':'float', 'is_vector':True, 'levels':level_sfc, 'units':'m/s'},
            'atmos_temp': {'name':'T', 'dtype':'float', 'is_vector':False, 'levels':levels, 'units':'K'},
            'atmos_pres': {'name':'P', 'dtype':'float', 'is_vector':False, 'levels':levels, 'units':'Pa'},
            'atmos_q_vapor': {'name':'QVAPOR', 'dtype':'float', 'is_vector':False, 'levels':levels, 'units':'kg/kg'},
            }

        self.run_process = None
        self.run_status = 'pending'

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        tstr = kwargs['time'].strftime('%Y-%m-%d_%H:%M:%S')
        return os.path.join(kwargs['path'], 'wrfout_'+tstr+'.nc')

    def read_grid(self, **kwargs):
        if self.map_proj == 'polar':
            proj = Proj(f'+proj=stere +lat_0={self.ref_lat} +lon_0={self.ref_lon} +lat_ts={self.truelat1}')

        elif self.map_proj == 'lambert':
            proj = Proj(f'+proj=lcc +lat_0={self.ref_lat} +lon_0={self.ref_lon} +lat_1={self.truelat1} +lat_2={self.truelat2}')

        elif self.map_proj == 'mercator':
            proj = Proj(f'+proj=merc +lat_0={self.ref_lat} +lon_0={self.ref_lon}')

        elif self.map_proj == 'lat-lon':
            proj = Proj(f'+proj=longlat')

        else:
            raise ValueError(f'unknown map_proj type {self.map_proj}')

        ##staggering here?
        ##again, use domain 1 settings for grid
        x_coords = (np.arange(self.e_we[0]) - self.ref_x) * self.dx
        y_coords = (np.arange(self.e_sn[0]) - self.ref_y) * self.dy
        x, y = np.meshgrid(x_coords, y_coords)

        self.grid = Grid(proj, x, y)

    def read_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        pass

    def write_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        pass

    def z_coords(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        pass

    def preprocess(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        pass

    def postprocess(self, task_id=0, **kwargs):
        pass

    def run(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        self.run_status = 'running'

        fname = self.filename(**kwargs)
        run_dir = os.path.dirname(fname)
        print('running wrf model in '+run_dir, flush=True)

        wrf_src = os.path.join(self.model_code_dir, 'setup.src')
        wrf_exe = os.path.join(self.model_code_dir, 'main', 'wrf.exe')

        log_file = 'rsl.error.0000'

        ##collect restart variables from bin and write to wrfrst

        ##build the run command
        shell_cmd = ". "+wrf_src+"; "   ##enter wrf env
        shell_cmd += f"JOB_EXECUTE {wrf_exe} >& run.log"

        run_job(shell_cmd, job_name='wrf.run', run_dir=run_dir,
                nproc=self.nproc_per_run, offset=task_id*self.nproc_per_run,
                walltime=self.walltime, **kwargs)

        # "SUCCESS COMPLETE" in log_file

        ##wrfrst at nexttime collect to bin file




import unittest
import importlib
from NEDAS.config import Config
from NEDAS.datasets import registry, get_dataset_class, Dataset

class TestDatasetInterface(unittest.TestCase):
    def test_dataset_class_init(self):
        c = Config()
        for dataset_name in registry.keys():
            DatasetClass = get_dataset_class(dataset_name)
            dataset = DatasetClass(grid=c.grid, mask=c.grid.mask)
            self.assertIsInstance(dataset, Dataset)

if __name__ == '__main__':
    unittest.main()



import numpy as np
from matplotlib.tri import Triangulation

def get_deformation_nodes(x, y, u, v):
    tri = Triangulation(x, y)
    e1, e2, e3, tri_a, tri_p = get_deformation_on_triangulation(x, y, u, v, tri.triangles)
    return e1, e2, e3, tri_a, tri_p, tri.triangles

def get_deformation_on_triangulation(x, y, u, v, t):
    xt, yt, ut, vt = [i[t].T for i in (x, y, u, v)]
    tri_x = np.diff(np.vstack([xt, xt[0]]), axis=0)
    tri_y = np.diff(np.vstack([yt, yt[0]]), axis=0)
    tri_s = np.hypot(tri_x, tri_y)
    tri_p = np.sum(tri_s, axis=0)
    s = tri_p/2
    tri_a = np.sqrt(s * (s - tri_s[0]) * (s - tri_s[1]) * (s - tri_s[2]))
    e1, e2, e3 = get_deformation_elems(xt, yt, ut, vt, tri_a)
    return e1, e2, e3, tri_a, tri_p

def get_deformation_elems(x, y, u, v, a):
    ux = uy = vx = vy = 0
    for i0, i1 in zip([1, 2, 0], [0, 1, 2]):
        ux += (u[i0] +u[i1]) * (y[i0] - y[i1])
        uy -= (u[i0] +u[i1]) * (x[i0] - x[i1])
        vx += (v[i0] +v[i1]) * (y[i0] - y[i1])
        vy += (v[i0] +v[i1]) * (x[i0] - x[i1])
    ux, uy, vx, vy = [i / (2 * a) for i in (ux, uy, vx, vy)]
    e1 = ux + vy
    e2 = np.hypot(ux - vy, uy + vx)
    e3 = vx - uy
    return e1, e2, e3



##from nextsim-tools/pynextsim

import numpy as np

##projection used in msh files
from pyproj import Proj
proj = Proj(proj='stere', a=6378273, b=6356889.448910593, lat_0=90., lon_0=-45., lat_ts=60.)

class MeshPhysicalName:
    """
    Class used to distinguish coastal or open boundaries

    MeshPhysicalName(name,ident,topodim)

    Parameters:
    -----------
    name : string
        eg 'coast' or 'open'
    ident : int
        eg edge number
    topodim : int
        eg 2 (2D space) or 3 (3D space)
    """
    def __init__(self, name, ident, topodim):
        self.name = name
        self.ident = ident
        self.topodim = topodim

    def write(self, fid):
        '''
        add a physical name to the output file

        Parameters:
        -----------
        fid : _io.TextIOWrapper
        '''
        fid.write('%i %i "%s"\n' %(
            self.topodim, self.ident, self.name))

class MeshElement:
    """ Class used for both individual triangles and edges """

    def __init__(self, ident, eltype, tags, node_ids, node_indices):
        """
        Parameters:
        -----------
        ident : int
            element number
        eltype : int
            determines if triangle (2) or edge (1)
        vertices : list(int)
            list of mesh node ids
        tags : list
            tags from .msh files
            physical, elementary...
        node_ids : list(int)
            list of node ids
        node_indices : list(int)
            list of node indices
        """
        self.ident = ident
        self.eltype = eltype
        self.node_ids = node_ids
        self.node_indices = node_indices
        self.tags = tags
        self.physical = None
        if len(tags)>0:
            self.physical = tags[0] #to distinguish open/coast
        self.num_vertices = len(node_ids)

    def get_coords(self, xnod, ynod):
        """
        clist = self.get_coords(xnod,ynod)

        Parameters:
        -----------
        xnod : np.ndarray
            x coords of nodes
        ynod : np.ndarray
            y coords of nodes

        Returns:
        --------
        clist : list
            list of tuples with x,y coords of nodes for the element
        """
        return [(xnod[n], ynod[n]) for n in self.node_indices]

    def write(self, fid):
        '''
        add the element info to the output file

        Parameters:
        -----------
        fid : _io.TextIOWrapper
        '''
        lst = ['%i %i %i' %(
            self.ident, self.eltype, len(self.tags))]
        lst += ['%i' %t for t in self.tags]
        lst += ['%i' %i for i in self.node_ids]
        fid.write(" ".join(lst) + "\n")

class GmshBoundary:
    """
    Class for handling mesh boundaries
    """
    def __init__(self, exterior, islands=None,
            open_boundaries=None, coastal_boundaries=None):
        """
        Parameters:
        -----------
        exterior : shapely.geometry.Polygon
        islands : list(shapely.geometry.Polygon)
            - internal closed boundaries
        open_boundaries : list (shapely.geometry.LineString)
            - external open boundaries
        coastal_boundaries : list (shapely.geometry.LineString)
            - external closed boundaries
        """

        self.exterior_polygon = exterior
        self.island_polygons = islands
        self.open_boundaries = open_boundaries
        self.coastal_boundaries = coastal_boundaries

        xe, ye = exterior.exterior.xy
        self._set_xy_range(xe, ye)
        self._set_resolution(xe, ye)

    def _set_xy_range(self, xe, ye):
        """
        Set the x-y range

        Parameters:
        -----------
        xe : numpy.ndarray
            x coords of exterior polygon
        ye : numpy.ndarray
            y coords of exterior polygon

        Sets:
        -----
        self.xmin : float
        self.xmax : float
        self.ymin : float
        self.ymax : float
        """
        self.xmin = np.min(xe)
        self.xmax = np.max(xe)
        self.ymin = np.min(ye)
        self.ymax = np.max(ye)

    def _set_resolution(self, xe, ye):
        """
        Estimate the mesh resolution

        Parameters:
        -----------
        xe : numpy.ndarray
            x coords of exterior polygon
        ye : numpy.ndarray
            y coords of exterior polygon

        Sets:
        -----
        self.resolution : float
        """
        dx = np.diff(xe)
        dy = np.diff(ye)
        self.resolution = np.mean(np.hypot(dx, dy))

    def iswet(self, x, y):
        """
        use matplotlib.path to test if multiple points are contained
        inside the polygon self.exterior_polygon

        Parameters:
        -----------
        x: numpy.ndarray
            x coordinates to test
        y: numpy.ndarray
            y coordinates to test

        Returns:
        --------
        wet : numpy.ndarray(bool)
            mask of same shape as x and y
            - element is True/False if corresponding point is inside/outside the mesh
              (inside external poly but outside island ones)
        """
        coords = np.array([x.flatten(), y.flatten()]).T
        # test if inside external polygon
        mask = self.points_in_polygon(self.exterior_polygon, coords)
        # test not inside island polygons
        for p in self.island_polygons:
            mask *= ~self.points_in_polygon(p, coords)
        return mask.reshape(x.shape)

def xyz_to_lonlat(x, y, z):
    radius = np.array(np.sqrt(pow(x, 2.)+pow(y, 2.)+pow(z, 2.)), dtype=np.float64)
    r2d = 180./np.pi
    lat = r2d * np.arcsin(z / radius)
    lon = r2d * np.arctan2(y, x)
    lon -= 360. * np.floor(lon / 360.) # 0 <= lon < 360
    return lon, lat

def lonlat_to_xyz(lon, lat):
    d2r = np.pi / 180.
    rlat = d2r * lat
    rlon = d2r * lon
    z = np.sin(rlat)
    r = np.cos(rlat)
    x = r * np.cos(rlon)
    y = r * np.sin(rlon)
    return x, y, z

###functions to handle msh files
def read_mshfile(filename):
    info = {}
    with open(filename, 'r') as f:
        version, fmt, size = read_version_info(f)
        physical_names = read_physical_names(f)
        nodes_x, nodes_y, nodes_id, nodes_map = read_nodes(f)
        num_elements, edges, triangles = read_elements(f, nodes_map)
        info.update({'version':version, 'fmt':fmt, 'size':size,
                     'physical_names':physical_names,
                     'nodes_x':nodes_x, 'nodes_y':nodes_y,
                     'nodes_id':nodes_id, 'nodes_map':nodes_map,
                     'num_elements':num_elements,
                     'edges':edges, 'triangles':triangles,
                     })
        return info

def read_version_info(f):
    if "$MeshFormat" not in f.readline():
        raise ValueError("expecting $MeshFormat - not found")
    version, fmt, size = f.readline().split()
    if "$EndMeshFormat" not in f.readline():
        raise ValueError("expecting $EndMeshFormat - not found")
    return np.float32(version), int(fmt), int(size)

def read_physical_names(f):
    # GMSH meshes can have Physical Names defined at start
    physical_names = {}
    num_physical_names = 0
    if "$PhysicalNames" not in f.readline():
        raise ValueError("expecting $PhysicalNames -  not found")
    num_physical_names = int(f.readline())
    for _ in range(num_physical_names):
        topodim, ident, name = f.readline().split()
        if name[0]=='"':
            name  = name[1:-1]
        physical_names.update({int(ident):
                    MeshPhysicalName(name, int(ident), int(topodim))})
    if "$EndPhysicalNames" not in f.readline():
        raise ValueError("expecting $EndPhysicalNames - not found")
    return physical_names

def read_nodes(f):
    lin = f.readline()
    nodstr = lin.split()[0]
    if not ( ("$NOD" in lin)
            or ("$Nodes" in lin)
            or ("$ParametricNodes" in lin) ):
        msg = ("""Invalid nodes string '%s' in gmsh importer.
                    It should be either '$NOD','$Nodes'
                    or '$ParametricNodes'."""
                    %(nodstr))
        raise ValueError(msg)

    num_nodes = int(f.readline())
    lines = [f.readline().strip() for n in range(num_nodes)]
    iccc = np.array([[np.float64(v) for v in line.split()] for line in lines])
    lon, lat = xyz_to_lonlat(iccc[:, 1], iccc[:, 2], iccc[:, 3])
    nodes_x, nodes_y = proj(lon, lat)
    nodes_id = iccc[:,0].astype(int)
    nodes_map = {nodes_id[i]: i for i in range(nodes_id.size)}

    if ("$End"+nodstr[1:] not in f.readline()):
        raise ValueError("expecting $End"+nodstr[1:] +" - not found")
    return nodes_x, nodes_y, nodes_id, nodes_map

def parse_element_string(lin, nodes_map):
    slin = lin.split()
    elnumber, eltype, num_tags = [int(s) for s in slin[:3]]
    tags = [int(s) for s in slin[3:3+num_tags]]
    node_ids = [int(s) for s in slin[3+num_tags:]]
    node_inds = [nodes_map[nid] for nid in node_ids]
    el_obj = MeshElement(elnumber, eltype, tags, node_ids, node_inds)
    return el_obj

def read_elements(f, nodes_map):
    if '$Elements' not in f.readline():
        raise ValueError("invalid elements string in gmsh importer.")

    num_elements = int(f.readline())
    edges = []
    triangles = []
    for _ in range(num_elements):
        el_obj = parse_element_string(f.readline(), nodes_map)
        if el_obj.num_vertices==3:
            triangles.append(el_obj)
        else:
            edges.append(el_obj)
    num_edges     = len(edges)
    num_triangles = len(triangles)

    if ("$EndElements" not in f.readline()):
        raise ValueError("expecting $EndElements - not found")
    return num_elements, edges, triangles

def write_mshfile(filename, mesh_info):
    print('Saving %s' %filename)
    with open(filename, 'w') as f:
        write_version_info(f, mesh_info)
        write_physical_names(f, mesh_info)
        write_nodes(f, mesh_info)
        write_elements(f, mesh_info)

def write_version_info(f, mesh_info):
    version = mesh_info['version']
    fmt = mesh_info['fmt']
    size = mesh_info['size']
    f.write("$MeshFormat\n")
    f.write("%3.1f %i %i\n" %(version, fmt, size))
    f.write("$EndMeshFormat\n")

def write_physical_names(f, mesh_info):
    pnames = mesh_info['physical_names']
    f.write("$PhysicalNames\n")
    f.write("%i\n" %len(pnames))
    for pn in pnames.values():
        pn.write(f)
    f.write("$EndPhysicalNames\n")

def write_nodes(f, mesh_info):
    lon, lat = proj(mesh_info['nodes_x'], mesh_info['nodes_y'], inverse=True)
    x, y, z = lonlat_to_xyz(lon, lat)
    f.write("$Nodes\n")
    f.write("%i\n" %len(x))
    for i, xi, yi, zi in zip(mesh_info['nodes_id'], x, y, z):
        f.write("%i %18.16f %18.16f %18.16f\n" %(i, xi, yi, zi))
    f.write("$EndNodes\n")

def write_elements(f, mesh_info):
    f.write("$Elements\n")
    f.write("%i\n" %mesh_info['num_elements'])
    for e in mesh_info['edges']: e.write(f)
    for t in mesh_info['triangles']: t.write(f)
    f.write("$EndElements\n")




import numpy as np
import unittest
from NEDAS.utils.conversion import units_convert

class TestUnitsConvert(unittest.TestCase):

    def test_units_convert_scalar(self):
        self.assertEqual(units_convert(100, 1, 1.), 0.01)
        self.assertEqual(units_convert('km', 'm', 1.), 1000.)
        self.assertEqual(units_convert('h', 's', 1.), 3600.)
        self.assertEqual(units_convert('m/s', 'km/h', 1.), 3.6)
        self.assertEqual(units_convert('kg', 'g', 1.), 1000.)
        self.assertEqual(units_convert('m/s', 'm/s', 1.), 1.)
        self.assertEqual(units_convert('C', 'K', 0.), 273.15)

    def test_units_convert_field(self):
        shape = (100, 100)
        self.assertEqual(np.mean(units_convert(100, 1, np.ones(shape))), 0.01)
        self.assertAlmostEqual(np.mean(units_convert('km', 'm', np.ones(shape))), 1000.)


if __name__ == '__main__':
    unittest.main()



from NEDAS.config import Config
from .base import State

registry = {
}

def get_state(c: Config) -> State:
    #if state_type not in registry.keys():
    #    return State
    #module = importlib.import_module('NEDAS.assim_tools.state.'+state_type)
    #StateClass = getattr(module, registry[state_type])
    #return StateClass
    return State(c)



import struct

def read_uf_data(file):
    data = []
    fmt_str = 'dd5sdddiiidddd?iidii'  ##according to mod_measurement type def
    ##value, variance, id_str, lon, lat, depth, ipiv, jpiv, ns, a1, a2, a3, a4, stat, i_orig_grid, j_orig_grid, h, date, orig_id
    recl = struct.calcsize(fmt_str)
    with open(file, 'rb') as f:
        while True:
            d = f.read(recl)
            if not d:
                break
            data.append(struct.unpack(fmt_str, d))
    return data




import numpy as np
from NEDAS.utils.njit import njit

@njit
def adjust_dp(dp, depth, onem):
    """
    Adjusts the pressure layers (dp) to ensure no negative values 
    and recalculates layer thickness based on depth constraints.

    Parameters:
    dp (ndarray): 3D array of pressure layers (kdm, jdm, idm).
    depth (ndarray): 2D array of depths (jdm, idm).
    onem (float): Scaling factor.

    Returns:
    ndarray: Adjusted dp array.
    """
    kdm, jdm, idm = dp.shape
    dp_ = dp.copy()  # Backup of the original dp
    press = np.zeros(kdm + 1)

    for j in range(jdm):
        for i in range(idm):
            # Move negative layers to neighbouring layers (forward sweep)
            for k in range(kdm - 1):
                dp[k+1, j, i] += min(0.0, dp[k, j, i])
                dp[k, j, i] = max(dp[k, j, i], 0.0)

            # Adjust lowermost layers (backward sweep)
            for k in range(kdm-1, 2, -1):
                dp[k-1, j, i] += min(0.0, dp[k, j, i])
                dp[k, j, i] = max(dp[k, j, i], 0.0)

            # Recalculate pressure layers
            press[0] = 0.0
            for k in range(kdm):
                press[k+1] = press[k] + dp[k, j, i]
                press[k+1] = min(depth[j, i] * onem, press[k+1])
            press[kdm] = depth[j, i] * onem

            # Recalculate dp based on updated pressures
            for k in range(kdm):
                dp[k, j, i] = press[k+1] - press[k]

            # Fallback for invalid depths
            if depth[j, i] > 100000.0 or depth[j, i] < 1.0:
                dp[:, j, i] = dp_[:, j, i]
    return dp

def stmt_fns_constans(thflag:int):
    if thflag == 0:
        # coefficients for sigma-0 (based on Brydon & Sun fit)
        c1 = -1.36471e-01
        c2 = 4.68181e-02
        c3 = 8.07004e-01
        c4 = -7.45353e-03
        c5 = -2.94418e-03
        c6 = 3.43570e-05
        c7 = 3.48658e-05
        pref = 0.0
    elif thflag == 2:
        # coefficients for sigma-2 (based on Brydon & Sun fit)
        c1 = 9.77093e+00
        c2 = -2.26493e-02
        c3 = 7.89879e-01
        c4 = -6.43205e-03
        c5 = -2.62983e-03
        c6 = 2.75835e-05
        c7 = 3.15235e-05
        pref = 2000.e4
    elif thflag == 4:
        # coefficients for sigma-4 (based on Brydon & Sun fit)
        c1 = 1.92362e+01
        c2 = -8.82080e-02
        c3 = 7.73552e-01
        c4 = -5.46858e-03
        c5 = -2.31866e-03
        c6 = 2.11306e-05
        c7 = 2.82474e-05
        pref = 4000.e4
    else:
        raise ValueError(f"Unknown thflag: {thflag}")
    return c1,c2,c3,c4,c5,c6,c7,pref

def stmt_fns_sigma(thflag:int, temp, saln):
    """Adapted from stmt_fns_sigma in EnKF-MPI-TOPAZ/SSHFromState"""
    c1,c2,c3,c4,c5,c6,c7,pref = stmt_fns_constans(thflag)
    return c1 + c3 * saln + temp * (c2 + c5 * saln + temp * (c4 + c7 * saln + c6 * temp))

def stmt_fns_kappaf(thflag:int, temp, saln, pres, thref=1.0e-3):
    c1,c2,c3,c4,c5,c6,c7,pref = stmt_fns_constans(thflag)
    # coefficients for kappa^(theta)
    # new values (w.r.t. t-toff, s-soff, prs-poff) from Shan Sun, 2/8/01
    toff = 3.0
    soff = 35.0
    poff = 2000.e4
    prefo = pref - poff
    qt = -2.41701E-01
    qs = -1.05131E-01
    qtt = 4.03001E-03
    qst = 1.06829E-03
    qttt = -3.03869E-05
    qpt = 1.00638E-09
    qpst = 1.48599E-11
    qptt = -1.31384E-11
    qthref = 1.0 / thref

    t = np.maximum(-2.0, np.minimum(32.0, temp)) - toff
    s = np.maximum(30.0, np.minimum(38.0, saln)) - soff
    prs = pres - poff

    kappaf = (1.0e-11 * qthref) * (prs - prefo) * (
              s * (qs + t * qst) +
              t * (qt + t * (qtt + t * qttt) +
              0.5 * (prs + prefo) * (qpt + s * qpst + t * qptt)) )
    return kappaf



import os
import numpy as np
from NEDAS.grid import Grid
from NEDAS.utils.conversion import dt1h
from NEDAS.utils.shell_utils import run_command, makedir
from NEDAS.utils.netcdf_lib import nc_read_var, nc_write_var
from NEDAS.models import Model
from .util import initial_condition, advance_time

class Vort2DModel(Model):
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        ##define the model grid
        ii, jj = np.meshgrid(np.arange(self.nx), np.arange(self.ny))
        x = ii*self.dx
        y = jj*self.dx
        self.grid = Grid(None, x, y, cyclic_dim='xy')
        self.grid.mask = np.full(self.grid.x.shape, False)  ##no mask

        levels = np.array([0])  ##there is no vertical levels

        self.variables = {'velocity': {'name':('u', 'v'), 'dtype':'float', 'is_vector':True, 'restart_dt':self.restart_dt, 'levels':levels, 'units':'m/s'}, }

        self.z_units = '*'

        self.run_process = None
        self.run_status = 'pending'

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        if kwargs['member'] is not None:
            mstr = '_mem{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''

        assert kwargs['time'] is not None, 'missing time in kwargs'
        tstr = kwargs['time'].strftime('%Y%m%d_%H')

        return os.path.join(kwargs['path'], tstr+mstr+'.nc')

    def read_grid(self, **kwargs):
        pass

    def read_mask(self, **kwargs):
        pass

    def read_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)

        rec = self.variables[kwargs['name']]
        comm = kwargs['comm']
        if rec['is_vector']:
            u = nc_read_var(fname, rec['name'][0], comm=comm)[0, ...]
            v = nc_read_var(fname, rec['name'][1], comm=comm)[0, ...]
            var = np.array([u, v])
        else:
            var = nc_read_var(fname, rec['name'], comm=comm)[0, ...]
        return var

    def write_var(self, var, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)

        rec = self.variables[kwargs['name']]
        comm = kwargs['comm']
        if rec['is_vector']:
            for i in range(2):
                nc_write_var(fname, {'t':None, 'y':self.ny, 'x':self.nx}, rec['name'][i], var[i,...], recno={'t':0}, comm=comm)
        else:
            nc_write_var(fname, {'t':None, 'y':self.ny, 'x':self.nx}, rec['name'], var, recno={'t':0}, comm=comm)

    def z_coords(self, **kwargs):
        return np.zeros(self.grid.x.shape)

    def generate_initial_condition(self):
        state = initial_condition(self.grid, self.Vmax, self.Rmw, self.Vbg, self.Vslope, self.loc_sprd)
        return state

    def preprocess(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        makedir(kwargs['path'])

        file1 = self.filename(**{**kwargs, 'path':kwargs['restart_dir']})
        file2 = self.filename(**kwargs)
        run_command(f"cp -fL {file1} {file2}")

    def postprocess(self, task_id=0, **kwargs):
        pass

    def run(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        self.run_status = 'running'

        makedir(kwargs['path'])

        state = self.read_var(**kwargs)
        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h
        next_state = advance_time(state, self.dx, forecast_period, self.dt, self.gen, self.diss)
        self.write_var(next_state, **{**kwargs, 'time':next_time})
        self.run_status = 'complete'



import os
import glob
from typing import Optional, Any
from functools import lru_cache
from datetime import datetime, timedelta, timezone
import numpy as np

from NEDAS.utils.conversion import units_convert, t2s, dt1h
from NEDAS.utils.netcdf_lib import nc_read_var, nc_write_var
from NEDAS.utils.shell_utils import run_command, run_job, makedir
from NEDAS.utils.progress import watch_log, find_keyword_in_file, watch_files
from NEDAS.models import Model

from ..time_format import dayfor
from ..abfile import ABFileRestart, ABFileArchv, ABFileForcing
from ..model_grid import get_topaz_grid, get_depth, get_mean_ssh, stagger, destagger
from .namelist import namelist
from .postproc import adjust_dp, stmt_fns_sigma, stmt_fns_kappaf
from .cice_utils import thickness_upper_limit, adjust_ice_variables, fix_zsin_profile

class Topaz5Model(Model):
    """
    TOPAZ5 model class.
    """
    nhc_root: str
    basedir: str
    model_env: Optional[str]
    reanalysis_code: str
    V: str
    X: str
    T: str
    R: str
    E: str
    idm: int
    jdm: int
    kdm: int
    nproc: int
    nproc_per_run: int
    nproc_per_util: int
    use_job_array: bool
    walltime: int
    stagnant_log_timeout: int
    ens_run_type: str
    meanssh_file: str
    forcing_file: str
    restart_dt: int
    output_dt: int
    forcing_dt: int
    z_units: str
    thflag: int
    thref: float
    thbase: float
    kapref: int
    yrflag: int
    ONEM: float
    MIN_SEAICE_CONC: float
    MAX_OCEAN_TEMP: float
    MIN_OCEAN_SALN: float
    MAX_OCEAN_SALN: float
    Nilayer: int
    saltmax: float
    min_salin: float
    depressT: float
    nsal: float
    msal: float
    aice_thresh: float
    fice_thresh: float
    hice_impact: float

    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        levels = np.arange(self.kdm) + 1  ##ocean levels, from top to bottom, k=1..kdm
        level_sfc = np.array([0])    ##some variables are only defined on surface level k=0
        level_ncat = np.arange(5)   ##some ice variables have 5 categories, treating them as levels also indexed by k

        self.restart_variables = {
            'ocean_velocity':    {'name':('u', 'v'), 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':levels, 'units':'m/s'},
            'ocean_layer_thick': {'name':'dp', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':levels, 'units':'Pa'},
            'ocean_temp':        {'name':'temp', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':levels, 'units':'C'},
            'ocean_saln':        {'name':'saln', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':levels, 'units':'psu'},
            'ocean_b_velocity':  {'name':('ubavg', 'vbavg'), 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'m/s'},
            'ocean_b_press':     {'name':'pbavg', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'Pa'},
            'ocean_mixl_depth':  {'name':'dpmixl', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'Pa'},
            'ocean_bot_press':   {'name':'pbot', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'Pa'},
            'ocean_bot_dense':   {'name':'thkk', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'?'},
            'ocean_bot_montg_pot': {'name':'psikk', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'?'},
            }

        self.archive_variables = {
            'ocean_velocity_daily': {'name':('u-vel.', 'v-vel.'), 'dtype':'float', 'is_vector':True, 'dt':self.output_dt, 'levels':levels, 'units':'m/s'},
            'ocean_layer_thick_daily': {'name':'thknss', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':levels, 'units':'Pa'},
            'ocean_temp_daily': {'name':'temp', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':levels, 'units':'C'},
            'ocean_saln_daily': {'name':'salin', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':levels, 'units':'psu'},
            'ocean_mixl_depth_daily': {'name':'mix_dpth', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':level_sfc, 'units':'Pa'},
            'ocean_dense_daily': {'name':'dense', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':level_sfc, 'units':'?'},
            'ocean_surf_height_daily': {'name':'srfhgt', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':level_sfc, 'units':'m'},
            }

        self.iced_variables = {
            'seaice_velocity': {'name':('uvel', 'vvel'), 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'m/s'},
            'seaice_conc_ncat':   {'name':'aicen', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_ncat, 'units':1},
            'seaice_volume_ncat':  {'name':'vicen', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_ncat, 'units':'m'},
            'snow_volume_ncat':  {'name':'vsnon', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_ncat, 'units':'m'},
            }

        self.iceh_variables = {
            'seaice_velocity_daily': {'name':('uvel_d', 'vvel_d'), 'dtype':'float', 'is_vector':True, 'dt':self.output_dt, 'levels':level_sfc, 'units':'m/s'},
            'seaice_conc_daily': {'name':'aice_d', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':level_sfc, 'units':1},
            'seaice_thick_daily': {'name':'hi_d', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':level_sfc, 'units':'m'},
            'seaice_surf_temp_daily': {'name':'Tsfc_d', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':level_sfc, 'units':'C'},
            'seaice_saln_daily': {'name':'sice_d', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':level_sfc, 'units':'psu'},
            'snow_thick_daily': {'name':'hs_d', 'dtype':'float', 'is_vector':False, 'dt':self.output_dt, 'levels':level_sfc, 'units':'m'},
            }

        self.atmos_forcing_variables = {
            'atmos_surf_velocity': {'name':('wndewd', 'wndnwd'), 'dtype':'float', 'is_vector':True, 'dt':self.forcing_dt, 'levels':level_sfc, 'units':'m/s'},
            'atmos_surf_temp':     {'name':'airtmp', 'dtype':'float', 'is_vector':False, 'dt':self.forcing_dt, 'levels':level_sfc, 'units':'C'},
            'atmos_surf_dewpoint': {'name':'dewpt', 'dtype':'float', 'is_vector':False, 'dt':self.forcing_dt, 'levels':level_sfc, 'units':'K'},
            'atmos_surf_press':    {'name':'mslprs', 'dtype':'float', 'is_vector':False, 'dt':self.forcing_dt, 'levels':level_sfc, 'units':'Pa'},
            'atmos_precip':        {'name':'precip', 'dtype':'float', 'is_vector':False, 'dt':self.forcing_dt, 'levels':level_sfc, 'units':'m/s'},
            'atmos_down_longwave': {'name':'radflx', 'dtype':'float', 'is_vector':False, 'dt':self.forcing_dt, 'levels':level_sfc, 'units':'W/m2'},
            'atmos_down_shortwave': {'name':'shwflx', 'dtype':'float', 'is_vector':False, 'dt':self.forcing_dt, 'levels':level_sfc, 'units':'W/m2'},
            'atmos_surf_vapor_mix': {'name':'vapmix', 'dtype':'float', 'is_vector':False, 'dt':self.forcing_dt, 'levels':level_sfc, 'units':'kg/kg'},
            }
        self.force_synoptic_names = [name for r in self.atmos_forcing_variables.values() for name in (r['name'] if isinstance(r['name'], tuple) else [r['name']])]

        self.diag_variables = {
            'ocean_surf_height': {'name':'ssh', 'operator':self.get_ocean_surf_height, 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'m'},
            'ocean_surf_height_anomaly': {'name':'sla', 'operator':self.get_ocean_surf_height_anomaly, 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'m'},
            'ocean_surf_temp': {'name':'sst', 'operator':self.get_ocean_surf_temp, 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'C'},
            'seaice_conc': {'name':'sic', 'operator':self.get_seaice_conc, 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':1},
            'seaice_thick': {'name':'sit', 'operator':self.get_seaice_thick, 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'m'},
            'snow_thick': {'name':'snwt', 'operator':self.get_snow_thick, 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':level_sfc, 'units':'m'},
            }

        self.variables = {**self.restart_variables,
                          **self.iced_variables,
                          **self.iceh_variables,
                          **self.atmos_forcing_variables,
                          **self.diag_variables,
                          **self.archive_variables}

        self.grid = None
        grid_info_file = os.path.join(self.basedir, 'topo', 'grid.info')
        if self.basedir and os.path.exists(grid_info_file):
            self.grid = get_topaz_grid(grid_info_file)

        self.depthfile = os.path.join(self.basedir, 'topo', f'depth_{self.R}_{self.T}.a')
        if self.grid and self.depthfile and os.path.exists(self.depthfile):
            self.depth, self.grid.mask = get_depth(self.depthfile, self.grid)

        self.meanssh = None
        if self.meanssh_file and os.path.exists(self.meanssh_file):
            self.meanssh = get_mean_ssh(self.meanssh_file, self.grid)

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        if kwargs['member'] is not None:
            mstr = '_mem{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''

        ##filename for each model component
        if kwargs['name'] in self.restart_variables:
            tstr = kwargs['time'].strftime('%Y_%j_%H_%M%S')
            file = 'restart.'+tstr+mstr+'.a'

        elif kwargs['name'] in self.iced_variables:
            t = kwargs['time']
            tstr = f"{t:%Y-%m-%d}-{t.hour*3600:05}"
            file = 'iced.'+tstr+mstr+'.nc'

        elif kwargs['name'] in self.iceh_variables:
            tstr = kwargs['time'].strftime('%Y-%m-%d')
            file = os.path.join(mstr[1:], 'SCRATCH', 'cice', 'iceh.'+tstr+'.nc')

        elif kwargs['name'] in self.atmos_forcing_variables:
            file = os.path.join(mstr[1:], 'SCRATCH', 'forcing')
            return os.path.join(kwargs['path'], file)

        elif kwargs['name'] in self.archive_variables:
            tstr = kwargs['time'].strftime('%Y_%j_??')  ##archive variables are daily means
            file = os.path.join(mstr[1:], 'SCRATCH', 'archm.'+tstr+'.a')

        elif kwargs['name'] in self.diag_variables:
            kstr = f"_k{kwargs['k']}_"
            tstr = t2s(kwargs['time'])
            file = os.path.join(mstr[1:], 'SCRATCH', kwargs['name']+kstr+tstr+'.npy')
            return os.path.join(kwargs['path'], file)

        else:
            raise ValueError(f"filename: ERROR: unknown variable name '{kwargs['name']}'")

        path = kwargs['path']
        fname = os.path.join(path, file)
        files = glob.glob(fname)
        if len(files) > 0:
            return files[0]

        ##if no corresponding files found under the given path
        ##try to find two layers above for the other cycle time
        dirs = path.split(os.sep)
        if dirs[-1] == 'topaz.v5' and len(dirs[-2])==12:
            root = os.sep.join(dirs[:-2])
            for tdir in os.listdir(root):
                fname = os.path.join(root, tdir, 'topaz.v5', file)
                files = glob.glob(fname)
                if len(files) > 0:
                    return files[0]
        raise FileNotFoundError(f"filename: ERROR: could not find {file} in {path} or its parent directories")

    def read_grid(self, **kwargs):
        pass

    def read_mask(self, **kwargs):
        pass

    def read_var(self, **kwargs):
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined")
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        rec = self.variables[name]

        ##get the variable from restart files
        if name in self.restart_variables:
            f = ABFileRestart(fname, 'r', idm=self.grid.nx, jdm=self.grid.ny)
            if rec['is_vector']:
                var1 = f.read_field(rec['name'][0], level=kwargs['k'], tlevel=1, mask=self.grid.mask)
                var2 = f.read_field(rec['name'][1], level=kwargs['k'], tlevel=1, mask=self.grid.mask)
                var = np.array([var1, var2])
            else:
                var = f.read_field(rec['name'], level=kwargs['k'], tlevel=1, mask=self.grid.mask)
            f.close()

        elif name in self.iced_variables:
            if rec['is_vector']:
                if name[-5:] == '_ncat':  ##ncat variable
                    var1 = nc_read_var(fname, rec['name'][0])[kwargs['k'],...]
                    var2 = nc_read_var(fname, rec['name'][1])[kwargs['k'],...]
                else:
                    var1 = nc_read_var(fname, rec['name'][0])
                    var2 = nc_read_var(fname, rec['name'][1])
                var = np.array([var1, var2])
            else:
                if name[-5:] == '_ncat':  ##ncat variable
                    var = nc_read_var(fname, rec['name'])[kwargs['k'],...]
                else:
                    var = nc_read_var(fname, rec['name'])

        elif name in self.iceh_variables:
            if rec['is_vector']:
                var1 = nc_read_var(fname, rec['name'][0])[0, ...]
                var2 = nc_read_var(fname, rec['name'][1])[0, ...]
                var = np.array([var1, var2])
            else:
                var = nc_read_var(fname, rec['name'])[0, ...]

        elif name in self.atmos_forcing_variables:
            dtime = (kwargs['time'] - datetime(1900,12,31,tzinfo=timezone.utc)) / (24*dt1h)
            if rec['is_vector']:
                f1 = ABFileForcing(fname+'.'+rec['name'][0], 'r')
                var1 = f1.read_field(rec['name'][0], dtime)
                f2 = ABFileForcing(fname+'.'+rec['name'][1], 'r')
                var2 = f2.read_field(rec['name'][1], dtime)
                var = np.array([var1, var2])
                f1.close()
                f2.close()
            else:
                f = ABFileForcing(fname+'.'+rec['name'], 'r')
                var = f.read_field(rec['name'], dtime)
                f.close()

        elif name in self.diag_variables:
            ## if the npy file exists, one could just read it to get the variable.
            ## but here we always calculate the variable from the model state, and refresh to the npy file, to be safe
            var = rec['operator'](**kwargs)
            np.save(fname, var)

        elif name in self.archive_variables:
            f = ABFileArchv(fname, 'r', mask=True)
            if rec['is_vector']:
                var1 = f.read_field(rec['name'][0], level=kwargs['k'])
                var2 = f.read_field(rec['name'][1], level=kwargs['k'])
                var = np.array([var1, var2])
            else:
                var = f.read_field(rec['name'], level=kwargs['k'])
            f.close()

        else:
            raise ValueError(f"read_var: ERROR: unknown variable name '{name}'")

        ##convert units if necessary
        var = units_convert(rec['units'], kwargs['units'], var)
        return var

    def write_var(self, var, **kwargs):
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined")
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        rec = self.variables[name]

        ##convert back to old units
        var = units_convert(kwargs['units'], rec['units'], var)

        if name in self.restart_variables:
            ##open the restart file for over-writing
            ##the 'r+' mode and a new overwrite_field method were added in the ABFileRestart in .abfile
            f = ABFileRestart(fname, 'r+', idm=self.grid.nx, jdm=self.grid.ny, mask=True)
            if rec['is_vector']:
                for i in range(2):
                    f.overwrite_field(var[i,...], self.grid.mask, rec['name'][i], level=kwargs['k'], tlevel=1)
            else:
                f.overwrite_field(var, self.grid.mask, rec['name'], level=kwargs['k'], tlevel=1)
            f.close()

        elif name in self.iced_variables:
            is_ncat = (name[-5:] == '_ncat')  ##if name is a multicategory variable (categories indexed by k)
            if rec['is_vector']:
                for i in range(2):
                    if is_ncat:
                        dims = {'ncat':None, 'nj':self.grid.ny, 'ni':self.grid.nx}
                        recno = {'ncat':kwargs['k']}
                    else:
                        dims = {'nj':self.grid.ny, 'ni':self.grid.nx}
                        recno = None
                    nc_write_var(fname, dims, rec['name'][i], var[i,...], recno=recno, comm=kwargs['comm'])
            else:
                if is_ncat:
                    dims = {'ncat':None, 'nj':self.grid.ny, 'ni':self.grid.nx}
                    recno = {'ncat':kwargs['k']}
                else:
                    dims = {'nj':self.grid.ny, 'ni':self.grid.nx}
                    recno = None
                nc_write_var(fname, dims, rec['name'], var, recno=recno, comm=kwargs['comm'])

        elif name in self.atmos_forcing_variables:
            dtime = (kwargs['time'] - datetime(1900,12,31,tzinfo=timezone.utc)) / timedelta(days=1)
            if rec['is_vector']:
                for i in range(2):
                    f = ABFileForcing(fname+'.'+rec['name'][i], 'r+')
                    f.overwrite_field(var[i,...], None, rec['name'][i], dtime)
                    f.close()
            else:
                f = ABFileForcing(fname+'.'+rec['name'], 'r+')
                f.overwrite_field(var, None, rec['name'], dtime)
                f.close()

        elif name in self.diag_variables:
            np.save(fname, var)

    def z_coords(self, **kwargs):
        return self._z_coords_cached(**kwargs)

    @lru_cache(maxsize=3)
    def _z_coords_cached(self, **kwargs):
        """
        Calculate vertical coordinates given the 3D model state
        Return:
        - z: np.array
        The corresponding z field
        """
        ##some defaults if not set in kwargs
        if 'k' not in kwargs:
            kwargs['k'] = 0

        z = np.zeros((self.jdm, self.idm))
        if kwargs['k'] == 0:
            ##if level index is 0, this is the surface, so just return zeros
            return z
        else:
            ##get layer thickness and convert to units
            rec = kwargs.copy()
            rec['name'] = 'ocean_layer_thick'
            rec['units'] = self.variables['ocean_layer_thick']['units'] ##should be Pa
            if self.z_units == 'm':
                dz = - self.read_var(**rec) / self.ONEM ##in meters, negative relative to surface
            elif self.z_units == 'Pa':
                dz = self.read_var(**rec)
            else:
                raise ValueError('do not know how to calculate z_coords for z_units = '+self.z_units)
            ##use recursive func, get previous layer z and add dz
            kwargs['k'] -= 1
            z_prev = self.z_coords(**kwargs)
            return z_prev + dz

    def get_ocean_surf_height(self, **kwargs):
        """Get ocean surface height from restart variables
        Adapted from p_ssh_from_state.F90 in ReanalysisTP5/SSHFromState_HYCOMICE"""
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined")
        tbaric = (self.kapref == self.thflag)

        restart_file = self.filename(**{**kwargs, 'name':'ocean_bot_montg_pot'})
        f = ABFileRestart(restart_file, 'r+', idm=self.grid.nx, jdm=self.grid.ny, mask=True)
        psikk = f.read_field('psikk', level=0, tlevel=1)
        thkk = f.read_field('thkk', level=0, tlevel=1)
        pbavg = f.read_field('pbavg', level=0, tlevel=1)

        ind = (self.depth < -0.1) & ~(np.isnan(self.depth))  ##valid points to calculate ssh on

        levels = list(self.variables['ocean_layer_thick']['levels'])
        idm, jdm, kdm = self.grid.nx, self.grid.ny, len(levels)
        pres = np.zeros((kdm+1, jdm, idm))     # cumulative pressure
        thstar = np.zeros((kdm, jdm, idm))
        montg = np.zeros((jdm, idm))
        oneta = np.zeros((jdm, idm))
        ssh = np.zeros((jdm, idm))

        for k in range(kdm):
            saln = f.read_field('saln', level=levels[k], tlevel=1)
            temp = f.read_field('temp', level=levels[k], tlevel=1)
            dp = f.read_field('dp', level=levels[k], tlevel=1)
            # use upper interface pressure in converting sigma to sigma-star
            # this is to avoid density variations in layers intersected by bottom
            th = stmt_fns_sigma(self.thflag, temp, saln)
            kapf = stmt_fns_kappaf(self.thflag, temp, saln, pres[k, ...], self.thref)
            if tbaric:
                thstar[k, ...][ind] = th[ind] + kapf[ind]
            else:
                thstar[k, ...][ind] = th[ind]
            pres[k+1, ...][ind] = pres[k, ...][ind] + dp[ind]
        oneta[ind] = 1. + pbavg[ind] / pres[-1, ...][ind]
        # m_prime in lowest layer
        montg[ind] = psikk[ind] + (pres[-1, ...][ind] * (thkk[ind] + self.thbase - thstar[-1, ...][ind]) - pbavg[ind] * (thstar[-1, ...][ind])) * self.thref**2
        # m_prime in remaining layers
        for k in range(kdm-2, -1, -1):
            montg[ind] += pres[k+1, ...][ind] * oneta[ind] * (thstar[k+1, ...][ind] - thstar[k, ...][ind]) * self.thref**2
        ssh[ind] = (montg[ind] / self.thref + pbavg[ind]) / self.ONEM
        ssh[~ind] = np.nan
        ssh[self.grid.mask] = np.nan
        return ssh

    def get_ocean_surf_height_anomaly(self, **kwargs):
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined")
        self.meanssh_file = os.path.join(self.basedir, 'topo', 'meanssh.uf')
        assert self.meanssh is not None, f"SLA: cannot find meanssh file {self.meanssh_file}"
        ssh = self.get_ocean_surf_height(**kwargs)
        sla = ssh - self.meanssh
        sla[self.grid.mask] = np.nan
        return sla

    def get_ocean_surf_temp(self, **kwargs):
        #just return first level ocean_temp
        return self.read_var(**{**kwargs, 'name':'ocean_temp', 'k':1})

    def get_seaice_conc(self, **kwargs):
        """
        Get total seaice concentration from multicategory ice concentration (aicen)
        adapted from ReanalysisTP5/SSHFromState_HYCOMICE/mod_read_icednc by J. Xie
        """
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined")
        seaice_conc = np.zeros(self.grid.x.shape)
        rec = kwargs.copy()
        rec['name'] = 'seaice_conc_ncat'
        rec['units'] = self.variables['seaice_conc_ncat']['units']
        for k in range(len(self.variables['seaice_conc_ncat']['levels'])):
            seaice_conc += self.read_var(**{**rec, 'k':k})
        
        seaice_conc[np.where(seaice_conc<self.MIN_SEAICE_CONC)] = 0.0  ##discard below threadshold
        seaice_conc[self.grid.mask] = np.nan
        return seaice_conc

    def get_seaice_thick(self, **kwargs):
        """
        Get total seaice thickness from the multicategory ice volume (vicen)
        """
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined before calling get_seaice_thick")
        seaice_conc = self.get_seaice_conc(**kwargs)
        
        seaice_volume = np.zeros(self.grid.x.shape)
        rec = kwargs.copy()
        rec['name'] = 'seaice_volume_ncat'
        rec['units'] = self.variables['seaice_volume_ncat']['units']
        for k in range(len(self.variables['seaice_volume_ncat']['levels'])):
            seaice_volume += self.read_var(**{**rec, 'k':k})
        
        seaice_thick = np.zeros(self.grid.x.shape)
        ind = np.where(seaice_conc>=self.MIN_SEAICE_CONC)
        upper_limit = thickness_upper_limit(seaice_conc[ind], 'seaice')
        seaice_thick[ind] = np.minimum(seaice_volume[ind] / seaice_conc[ind], upper_limit)
        seaice_thick[self.grid.mask] = np.nan
        return seaice_thick

    def get_snow_thick(self, **kwargs):
        """
        Get total snow thickness from the multi-category snow volume (vsnon)
        """
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined before calling get_snow_thick")
        seaice_conc = self.get_seaice_conc(**kwargs)

        snow_volume = np.zeros(self.grid.x.shape)
        rec = kwargs.copy()
        rec['name'] = 'snow_volume_ncat'
        rec['units'] = self.variables['snow_volume_ncat']['units']
        for k in range(len(self.variables['snow_volume_ncat']['levels'])):
            snow_volume += self.read_var(**{**rec, 'k':k})
        
        snow_thick = np.zeros(self.grid.x.shape)
        ind = np.where(seaice_conc>=self.MIN_SEAICE_CONC)
        upper_limit = thickness_upper_limit(seaice_conc[ind], 'snow')
        snow_thick[ind] = np.minimum(snow_volume[ind] / seaice_conc[ind], upper_limit)
        return snow_thick

    def preprocess(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        offset = task_id * self.nproc_per_util
        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h

        if kwargs['member'] is not None:
            mstr = '_mem{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''
        run_dir = os.path.join(kwargs['path'], mstr[1:], 'SCRATCH')
        ##make sure model run directory exists
        makedir(run_dir)

        ##generate namelists, blkdat, ice_in, etc.
        namelist(self, time, forecast_period, run_dir)

        ##copy synoptic forcing fields from a long record in basedir, will be perturbed later
        for varname in self.force_synoptic_names:
            forcing_file = self.forcing_file.format(member=kwargs['member']+1, time=time, name=varname)
            forcing_file_out = os.path.join(run_dir, 'forcing.'+varname)
            f = ABFileForcing(forcing_file, 'r')
            fo = ABFileForcing(forcing_file_out, 'w', idm=f.idm, jdm=f.jdm, cline1=f._cline1, cline2=f._cline2)
            t = time
            dt = self.forcing_dt
            rdtime = dt / 24
            while t <= next_time:
                dtime1 = dayfor(self.yrflag, t.year, int(t.strftime('%j')), t.hour)
                fld = f.read_field(varname, dtime1)
                fo.write_field(fld, None, varname, dtime1, rdtime)
                t += dt * dt1h
            f.close()
            fo.close()

        ##link necessary files for model run
        shell_cmd = f"cd {run_dir}; "
        ##partition setting
        partit_file = os.path.join(self.basedir, 'topo', 'partit', f'depth_{self.R}_{self.T}.{self.nproc:04d}')
        shell_cmd += f"ln -fs {partit_file} patch.input; "
        ##topo files
        for ext in ['.a', '.b']:
            file = os.path.join(self.basedir, 'topo', 'regional.grid'+ext)
            shell_cmd += f"ln -fs {file} .; "
            file = os.path.join(self.basedir, 'topo', f'depth_{self.R}_{self.T}'+ext)
            shell_cmd += f"ln -fs {file} regional.depth{ext}; "
        file = os.path.join(self.basedir, 'topo', f'kmt_{self.R}_{self.T}.nc')
        shell_cmd += f"ln -fs {file} cice_kmt.nc; "
        file = os.path.join(self.basedir, 'topo', 'cice_grid.nc')
        shell_cmd += f"ln -fs {file} .; "
        ##nest files
        nest_dir = os.path.join(self.basedir, 'nest', self.E)
        shell_cmd += f"ln -fs {nest_dir} nest; "
        ##TODO: there is extra logic in nhc_root/bin/expt_preprocess.sh to be added here
        ##relax files
        for ext in ['.a', '.b']:
            for varname in ['intf', 'saln', 'temp']:
                file = os.path.join(self.basedir, 'relax', self.E, 'relax_'+varname[:3]+ext)
                shell_cmd += f"ln -fs {file} {'relax.'+varname+ext}; "
            for varname in ['thkdf4', 'veldf4']:
                file = os.path.join(self.basedir, 'relax', self.E, varname+ext)
                shell_cmd += f"ln -fs {file} {varname+ext}; "
        ##other forcing files
        for ext in ['.a', '.b']:
            ##rivers
            file = os.path.join(self.basedir, 'force', 'rivers', self.E, 'rivers'+ext)
            shell_cmd += f"ln -fs {file} {'forcing.rivers'+ext}; "
            ##seawifs
            file = os.path.join(self.basedir, 'force', 'seawifs', 'kpar'+ext)
            shell_cmd += f"ln -fs {file} {'forcing.kpar'+ext}; "
        run_command(shell_cmd)

        ##copy restart files from restart_dir
        restart_dir = kwargs['restart_dir']
        ##job_submit_cmd = kwargs['job_submit_cmd']
        tstr = time.strftime('%Y_%j_%H_%M%S')
        for ext in ['.a', '.b']:
            file = os.path.join(restart_dir, 'restart.'+tstr+mstr+ext)
            file1 = os.path.join(kwargs['path'], 'restart.'+tstr+mstr+ext)
            run_command(f"cp -fL {file} {file1}")
            run_command(f"ln -fs {file1} {os.path.join(run_dir, 'restart.'+tstr+ext)}")
        makedir(os.path.join(run_dir, 'cice'))
        tstr = f"{time:%Y-%m-%d}-{time.hour*3600:05}"
        file = os.path.join(restart_dir, 'iced.'+tstr+mstr+'.nc')
        file1 = os.path.join(kwargs['path'], 'iced.'+tstr+mstr+'.nc')
        run_command(f"cp -fL {file} {file1}")
        run_command(f"ln -fs {file1} {os.path.join(run_dir, 'cice', 'iced.'+tstr+'.nc')}")
        run_command(f"echo {os.path.join('.', 'cice', 'iced.'+tstr+'.nc')} > {os.path.join(run_dir, 'cice', 'ice.restart_file')}")

    def postprocess(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined")
        time = kwargs['time']
        member = kwargs['member']
        if member is not None:
            mstr = '_mem{:03d}'.format(member+1)
        else:
            mstr = ''
        run_dir = os.path.join(kwargs['path'], mstr[1:], 'SCRATCH')
        makedir(run_dir)

        # link files
        commands = ""
        if self.model_env:
            commands += f". {self.model_env}; "
        commands += f"cd {run_dir}; "
        for ext in ['.a', '.b']:
            file1 = os.path.join(kwargs['restart_dir'], f'restart.{time:%Y_%j_%H_%M%S}{mstr}{ext}')
            file2 = f'forecast{member+1:03}{ext}'
            commands += f"ln -fs {file1} {file2}; "
        file1 = os.path.join(kwargs['restart_dir'], f'iced.{time:%Y-%m-%d}-{time.hour*3600:05}{mstr}.nc')
        file2 = f'ice_forecast{member+1:03}.nc'
        commands += f"ln -fs {file1} {file2}; "
        commands += f"ln -fs {self.reanalysis_code}/FILES/depths{self.idm}x{self.jdm}.uf .; "
        run_command(commands)

        # restart2nc on forecast files
        commands = ""
        if self.model_env:
            commands += f". {self.model_env}; "
        commands += f"cd {run_dir}; "
        commands += f"{os.path.join(self.reanalysis_code, 'ASSIM', 'BIN', 'restart2nc')} forecast{member+1:03}.a ice_forecast{member+1:03}.nc"
        run_command(commands)

        # add posterior ice variables in analysis abfile
        for ext in ['.a', '.b']:
            file1 = os.path.join(kwargs['path'], f'restart.{time:%Y_%j_%H_%M%S}{mstr}{ext}')
            file2 = os.path.join(run_dir, f'analysis{member+1:03}{ext}')
            run_command(f"cp -L {file1} {file2}")
        f = ABFileRestart(file2, 'r+', idm=self.grid.nx, jdm=self.grid.ny, mask=True)
        nfld = len(f.fields.keys())
        fld = np.load(self.filename(path=kwargs['path'], name='seaice_conc', member=member, time=time))
        f.write_field(fld, None, 'ficem', 0, 1, nfld)
        fld = np.load(self.filename(path=kwargs['path'], name='seaice_thick', member=member, time=time))
        f.write_field(fld, None, 'hicem', 0, 1, nfld+1)
        f.close()

        # run fixhycom and update restart file
        commands = ""
        if self.model_env:
            commands += f". {self.model_env}; "
        commands += f"cd {run_dir}; "
        commands += f"{os.path.join(self.reanalysis_code, 'ASSIM', 'BIN', 'fixhycom')} analysis{member+1:03}.a {member+1} forecast{member+1:03}.nc ice_forecast{member+1:03}.nc {time:%j} 0; "
        commands += f"cat fixanalysis{member+1:03}.b >> tmp{member+1:03}.b; mv tmp{member+1:03}.b fixanalysis{member+1:03}.b; "
        for ext in ['.a', '.b']:
            file1 = os.path.join(run_dir, f'fixanalysis{member+1:03}{ext}')
            file2 = os.path.join(kwargs['path'], f'restart.{time:%Y_%j_%H_%M%S}{mstr}{ext}')
            commands += f"mv {file1} {file2}; "
        file1 = os.path.join(run_dir, f'fix_ice_forecast{member+1:03}.nc')
        file2 = os.path.join(kwargs['path'], f'iced.{time:%Y-%m-%d}-{time.hour*3600:05}{mstr}.nc')
        commands += f"mv {file1} {file2}; "
        run_command(commands)

    def postprocess_native(self, task_id=0, **kwargs):
        """Post processing the restart variables for next forecast"""
        ## routines adapted from the EnKF-MPI-TOPAZ/Tools/fixhycom.F90 code
        kwargs = super().parse_kwargs(**kwargs)
        if self.grid is None:
            raise AttributeError("topaz5model: grid not yet defined")

        ##adjust ocean layer thickness dp
        rec = kwargs.copy()
        rec['name'] = 'ocean_layer_thick'
        rec['units'] = self.variables['ocean_layer_thick']['units'] ##should be Pa
        levels = list(self.variables['ocean_layer_thick']['levels'])
        dp = np.zeros((len(levels), self.grid.ny, self.grid.nx))
        for ilev, k in enumerate(levels):
            dp[ilev, ...] = self.read_var(**{**rec, 'k':k})
        dp = adjust_dp(dp, -self.depth, self.ONEM)
        for ilev, k in enumerate(levels):
            self.write_var(dp[ilev,...], **{**rec, 'k':k})

        ##loop over fields in restart file
        restart_file = self.filename(**{**kwargs, 'name':'ocean_temp'})
        f = ABFileRestart(restart_file, 'r+', idm=self.grid.nx, jdm=self.grid.ny, mask=True)
        for i, rec in f.fields.items():
            name = rec['field']
            tlevel = rec['tlevel']
            k = rec['k']
            fld = f.read_field(name, tlevel=tlevel, level=k)

            ##reset variables out of their normal range
            if name == 'temp':
                saln = f.read_field('saln', tlevel=tlevel, level=k)
                temp_min = -0.057 * saln
                ind = np.where(fld < temp_min)
                fld[ind] = temp_min[ind]
                fld[np.where(fld > self.MAX_OCEAN_TEMP)] = self.MAX_OCEAN_TEMP
            elif name == 'saln':
                fld[np.where(fld > self.MAX_OCEAN_SALN)] = self.MAX_OCEAN_SALN
                fld[np.where(fld < self.MIN_OCEAN_SALN)] = self.MIN_OCEAN_SALN
            elif name == 'dp':
                fld = dp[levels.index(k), ...]  ##set dp to the adjusted value

            ##write the field back
            f.overwrite_field(fld, None, name, tlevel=tlevel, level=k)
        f.close()

        ##fix sea ice variables, from enkf-topaz/Tools/m_put_mod_fld_nc: fix_cice
        restart_dir = kwargs['restart_dir']
        prior_ice_file = self.filename(**{**kwargs, 'path':restart_dir, 'name':'seaice_conc_ncat'})
        post_ice_file = self.filename(**{**kwargs, 'name':'seaice_conc_ncat'})
        fice = self.read_var(**{**kwargs, 'name':'seaice_conc', 'k':0, 'units':1})
        hice = self.read_var(**{**kwargs, 'name':'seaice_thick', 'k':0, 'units':'m'})
        zSin, Tmlt = fix_zsin_profile(self.Nilayer+1, self.saltmax, self.depressT, self.nsal, self.msal)
        adjust_ice_variables(prior_ice_file, post_ice_file, fice, hice, self.grid.mask,
                             self.aice_thresh, self.fice_thresh, self.hice_impact, zSin, Tmlt)

        ##update the diagnostic ice variables
        self.write_var(fice, **{**kwargs, 'name':'seaice_conc', 'k':0, 'units':1})
        self.write_var(hice, **{**kwargs, 'name':'seaice_thick', 'k':0, 'units':'m'})

    def run(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        self.run_status = 'running'

        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h

        member = kwargs['member']
        if member is not None:
            mstr = '_mem{:03d}'.format(member+1)
        else:
            mstr = ''
        run_dir = os.path.join(kwargs['path'], mstr[1:], 'SCRATCH')
        makedir(run_dir)
        log_file = os.path.join(run_dir, "run.log")
        run_command("touch "+log_file)

        run_success = False
        ##early exit if the run is already finished
        if find_keyword_in_file(log_file, 'Exiting hycom_cice'):
            run_success = True

        else:
            ##check if input file exists
            input_files = []
            tstr = time.strftime('%Y_%j_%H_%M%S')
            for ext in ['.a', '.b']:
                input_files.append(os.path.join(run_dir, 'restart.'+tstr+ext))
            tstr = f"{time:%Y-%m-%d}-{time.hour*3600:05}"
            input_files.append(os.path.join(run_dir, 'cice', 'iced.'+tstr+'.nc'))
            for file in input_files:
                if not os.path.exists(file):
                    raise RuntimeError(f"topaz.v5.model.run: input file missing: {file}")

            ##clean up some files from previous runs
            run_command(f"cd {run_dir}; rm -f archm.* ovrtn_out summary_out")
            run_command(f"echo > {log_file}")

            ##build the shell command line
            model_exe = os.path.join(self.basedir, f'expt_{self.X}', 'build', f'src_{self.V}ZA-07Tsig0-i-sm-sse_relo_mpi', 'hycom_cice')
            shell_cmd = ""
            if self.model_env:
                shell_cmd =  ". "+self.model_env+"; "  ##enter topaz5 env
            shell_cmd += "cd "+run_dir+"; "             ##enter run directory
            shell_cmd += 'JOB_EXECUTE '+model_exe+" >& run.log"

            ##run the model, give it 3 attempts
            for i in range(3):
                try:
                    run_job(shell_cmd, job_name='topaz5', run_dir=run_dir,
                            nproc=self.nproc, offset=task_id*self.nproc_per_run,
                            walltime=self.walltime, log_file=log_file, **kwargs)
                except RuntimeError as e:
                    print(f"{e}, retrying ({2-i} attempts remain)")
                    run_command(f"cp {log_file} {log_file}.attempt{i}")
                    continue
                ##check output
                if find_keyword_in_file(log_file, 'Exiting hycom_cice'):
                    run_success = True
                    break
        assert run_success, f"model run failed after 3 attempts, check in {run_dir}"

        ##move the output restart files to forecast_dir
        tstr = next_time.strftime('%Y_%j_%H_%M%S')
        for ext in ['.a', '.b']:
            file1 = os.path.join(run_dir, 'restart.'+tstr+ext)
            file2 = os.path.join(kwargs['path'], 'restart.'+tstr+mstr+ext)
            if os.path.exists(file1):
                run_command(f"mv {file1} {file2}")
            else:
                assert os.path.exists(file2), f"error moving output file {file1} to {file2}"
        tstr = f"{next_time:%Y-%m-%d}-{next_time.hour*3600:05}"
        file1 = os.path.join(run_dir, 'cice', 'iced.'+tstr+'.nc')
        file2 = os.path.join(kwargs['path'], 'iced.'+tstr+mstr+'.nc')
        if os.path.exists(file1):
            run_command(f"mv {file1} {file2}")
        else:
            assert os.path.exists(file2), f"error moving output file {file1} to {file2}"

    def run_batch(self, **kwargs):
        raise NotImplementedError("topaz5model.run_batch: not implemented, use run() instead")



import importlib
from typing import Type, TYPE_CHECKING
if TYPE_CHECKING:
    from NEDAS.datasets import Dataset

registry = {
    'lorenz96': 'Lorenz96Obs',
    'qg': 'QGObs',
    'vort2d': 'Vort2DObs',
    'ecmwf.era5': 'ERA5Data',
    'ecmwf.forecast': 'EcmwfForecastData',
    'ifremer.argo': 'ArgoObs',
    'osisaf.ice_conc': 'OsisafSeaIceConcObs',
    'osisaf.ice_drift': 'OsisafSeaIceDriftObs',
    'cs2smos': 'Cs2SmosObs',
    'rgps': 'RgpsObs',
    'topaz': 'TopazPrepObs',
}

def get_dataset_class(dataset_name: str) -> Type["Dataset"]:
    """
    Factory function to return the correct Dataset subclass.

    Args:
        dataset_name (str): Dataset name

    Returns:
        Type[Dataset]: Corresponding Dataset subclass.
    """
    dataset_name = dataset_name.lower()

    if dataset_name not in registry.keys():
        raise NotImplementedError(f"Dataset class not implemented for '{dataset_name}'")

    module = importlib.import_module('NEDAS.datasets.'+dataset_name)
    DatasetClass = getattr(module, registry[dataset_name])

    return DatasetClass



import numpy as np
import os
import subprocess
from functools import lru_cache

from NEDAS.utils.conversion import units_convert, dt1h
from NEDAS.utils.shell_utils import run_job
from NEDAS.models import Model
from ..abfile import ABFileRestart, ABFileBathy
from ..model_grid import get_topaz_grid
from .namelist import namelist

class Topaz4Model(Model):
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        levels = np.arange(1, 51, 1)
        level_sfc = np.array([0])
        self.variables = {
            'ocean_velocity': {'name':('u', 'v'), 'dtype':'float', 'is_vector':True, 'levels':levels, 'units':'m/s'},
            'ocean_layer_thick': {'name':'dp', 'dtype':'float', 'is_vector':False, 'levels':levels, 'units':'Pa'},
            'ocean_temp': {'name':'temp', 'dtype':'float', 'is_vector':False, 'levels':levels, 'units':'K'},
            'ocean_saln': {'name':'saln', 'dtype':'float', 'is_vector':False, 'levels':levels, 'units':'psu'},
            'ocean_surf_height': {'name':'msshb', 'dtype':'float', 'is_vector':False, 'levels':[0], 'units':'m'},
            'ocean_surf_temp': {'name':'sstb', 'dtype':'float', 'is_vector':False, 'levels':[0], 'units':'K'},
            'ocean_b_velocity':  {'name':('ubavg', 'vbavg'), 'dtype':'float', 'is_vector':True, 'levels':level_sfc, 'units':'m/s'},
            'ocean_b_press': {'name':'pbavg', 'dtype':'float', 'is_vector':False, 'levels':level_sfc, 'units':'Pa'},
            'ocean_mixl_depth': {'name':'dpmixl', 'dtype':'float', 'is_vector':False, 'levels':level_sfc, 'units':'Pa'},
            }

        self.z_units = 'm'
        
        ##model grid
        grid_info_file = os.path.join(self.basedir, 'topo', 'grid.info')
        self.grid = get_topaz_grid(grid_info_file)

        self.depthfile = os.path.join(self.basedir, 'topo', f'depth_{self.R}_{self.T}.a')
        f = ABFileBathy(self.depthfile, 'r', idm=self.grid.nx, jdm=self.grid.ny)
        depth = f.read_field('depth')
        f.close()
        self.depth = -depth.data
        self.grid.mask = depth.mask
        
    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        if kwargs['member'] is not None:
            mstr = '_mem{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''
        tstr = kwargs['time'].strftime('%Y_%j_%H_0000')
        return os.path.join(kwargs['path'], mstr[1:], 'TP4restart'+tstr+mstr+'.a')

    def read_grid(self, **kwargs):
        pass

    def read_mask(self):
        pass

    def read_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        rec = self.variables[name]

        f = ABFileRestart(fname, 'r', idm=self.grid.nx, jdm=self.grid.ny)
        if rec['is_vector']:
            var1 = f.read_field(rec[name]['name'][0], level=kwargs['k'], tlevel=1, mask=None)
            var2 = f.read_field(rec[name]['name'][1], level=kwargs['k'], tlevel=1, mask=None)
            var = np.array([var1, var2])
        else:
            var = f.read_field(rec[name]['name'], level=kwargs['k'], tlevel=1, mask=None)
        f.close()

        var = units_convert(rec['units'], kwargs['units'], var)
        return var

    def write_var(self, var, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        rec = self.variables[name]

        ##open the restart file for over-writing
        ##the 'r+' mode and a new overwrite_field method were added in the ABFileRestart in .abfile
        f = ABFileRestart(fname, 'r+', idm=self.grid.nx, jdm=self.grid.ny)

        ##convert units back if necessary
        var = units_convert(kwargs['units'], rec['units'], var)

        if rec['is_vector']:
            for i in range(2):
                f.overwrite_field(var[i,...], None, rec['name'][i], level=kwargs['k'], tlevel=1)
        else:
            f.overwrite_field(var, None, rec['name'], level=kwargs['k'], tlevel=1)
        f.close()

    @lru_cache(maxsize=3)
    def z_coords(self, **kwargs):
        """calculate vertical coordinates given the 3D model state
        """
        """
        Calculate vertical coordinates given the 3D model state
        Return:
        - z: np.array
        The corresponding z field
        """
        ##some defaults if not set in kwargs
        if 'k' not in kwargs:
            kwargs['k'] = 0

        z = np.zeros(self.grid.x.shape)

        if kwargs['k'] == 0:
            ##if level index is 0, this is the surface, so just return zeros
            return z

        else:
            ##get layer thickness and convert to units
            rec = kwargs.copy()
            rec['name'] = 'ocean_layer_thick'
            rec['units'] = self.variables['ocean_layer_thick']['units'] ##should be Pa
            if self.z_units == 'm':
                dz = - self.read_var(**rec) / self.onem ##in meters, negative relative to surface
            elif self.z_units == 'Pa':
                dz = self.read_var(**rec)
            else:
                raise ValueError('do not know how to calculate z_coords for z_units = '+self.z_units)

            ##use recursive func, get previous layer z and add dz
            kwargs['k'] -= 1
            z_prev = self.z_coords(**kwargs)
            return z_prev + dz

    def preprocess(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        init_file = self.self.filename(**{**kwargs, 'path':self.ens_init_dir})
        input_file = self.self.filename(**kwargs)
        os.system("mkdir -p "+os.path.dirname(input_file))
        os.system("cp "+init_file+" "+input_file)
        os.system("cp "+init_file.replace('.a', '.b')+" "+input_file.replace('.a', '.b'))

    def postprocess(self, task_id=0, **kwargs):
        pass

    def run(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        self.run_status = 'running'

        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h

        input_file = self.self.filename(**kwargs)
        run_dir = os.path.dirname(input_file)
        os.system("mkdir -p "+run_dir)
        os.chdir(run_dir)
        log_file = os.path.join(run_dir, "run.log")
        os.system('touch '+log_file)

        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h

        kwargs_out = {**kwargs, 'time':next_time}
        output_file = self.self.filename(**kwargs_out)

        ##create namelist config files
        namelist(self, time, forecast_period, run_dir)

        ##link files
        partit_file = os.path.join(self.basedir, 'topo', 'partit', f'depth_{self.R}_{self.T}.{self.nproc:04d}')
        os.system("cp "+partit_file+" patch.input")

        for ext in ['.a', '.b']:
            os.system("ln -fs "+os.path.join(self.basedir, 'topo', 'regional.grid'+ext)+" regional.grid"+ext)
            os.system("ln -fs "+os.path.join(self.basedir, 'topo', f'depth_{self.R}_{self.T}'+ext)+" regional.depth"+ext)
            os.system("ln -fs "+os.path.join(self.basedir, 'topo', 'tbaric'+ext)+" tbaric"+ext)
        os.system("ln -fs "+os.path.join(self.basedir, 'topo', 'grid.info')+" grid.info")

        ##TODO: switches for other forcing options
        if self.forcing_frc == 'era5':
            forcing_path = self.era5_path
        if self.forcing_frc == 'era40':
            pass
        os.system("ln -fs "+forcing_path+" .")
        os.system("ln -fs "+os.path.join(self.basedir, 'force', 'other', 'iwh_tabulated.dat')+" .")
        for ext in ['.a', '.b']:
            if self.priver == 1:
                os.system("ln -fs "+os.path.join(self.basedir, 'force', 'rivers', self.E, 'rivers'+ext)+" forcing.rivers"+ext)
            if self.jerlv0 == 0:
                os.system("ln -fs "+os.path.join(self.basedir, 'force', 'seawifs', 'kpar'+ext)+" forcing.kpar"+ext)
            if self.relax == 1:
                for comp in ['saln', 'temp', 'intf', 'rmu']:
                    os.system("ln -fs "+os.path.join(self.basedir, 'relax', self.E, 'relax_'+comp[:3]+ext)+" relax."+comp+ext)
            os.system("ln -fs "+os.path.join(self.basedir, 'relax', self.E, 'thkdf4'+ext)+" thkdf4"+ext)
        os.system("ln -fs "+os.path.join(self.basedir, 'relax', self.E, 'clim_tran.txt')+" .")
        # if self.gpflag: TODO
        # if self.nestoflag:
        # if self.nestiflag:
        # if self.tideflag:

        model_src = os.path.join(self.basedir, 'setup.src')
        model_exe = os.path.join(self.basedir, f'Build_V{self.V}_X{self.X}', 'hycom')

        ##build the shell command line
        shell_cmd =  ". "+model_src+"; "   ##enter topaz v4 env
        shell_cmd += "cd "+run_dir+"; "          ##enter run directory
        shell_cmd += f"JOB_EXECUTE {model_exe} {kwargs['member']+1} >& run.log"

        for tr in range(2):  ##number of tries
            with open(log_file, 'rt') as f:
                if '(normal)' in f.read():
                    break
            self.run_process = subprocess.Popen(shell_cmd, shell=True)
            self.run_process.wait()
            run_job(shell_cmd, job_name='topaz4_run', run_dir=run_dir,
                    nproc=self.nproc, offset=task_id*self.nproc_per_run,
                    walltime=self.walltime, **kwargs)

        with open(log_file, 'rt') as f:
            if '(normal)' not in f.read():
                raise RuntimeError('errors in '+log_file)
        if not os.path.exists(output_file):
            raise RuntimeError('output file not found: '+output_file)



import numpy
import datetime

_stefanb=5.67e-8
_s0=1365.                 # w/m^2  solar constant
_absh2o=0.09              # ---    absorption of water and ozone
_airdns=1.2

def lwrad_berliand(tair,e,cc) :
   # downwelling longwave radiation
   #tair : air temperature [K]
   #e    : near surface vapor pressure [Pa]
   #cc   : cloud cover (0-1)

   # Below formula assumes pressure in mBar
   e_mbar = e * 0.01
   emiss=0.97
   # Clear sky downwelling longwave flux from Eimova(1961)
   strd = emiss*_stefanb * tair**4 * (0.39-0.05*numpy.sqrt(e_mbar))

   # Cloud correction by Berliand (1952)
   strd = strd * (1. - 0.6823 * cc*cc)

   return strd


def strd_bignami(tair,e,cc) :
   # downwelling longwave radiation
   #tair : air temperature [K]
   #e    : near surface vapor pressure [Pa]
   #cc   : cloud cover (0-1)

   # Below formula assumes pressure in mBar (1hPa=1 mbar; 1Pas=0.01mbar)
   e_mbar = e * 0.01
   emiss=0.97
   # Clear sky downwelling longwave flux from Bignami(1995)
   strd = _stefanb * tair**4 * (0.653+0.00535*e_mbar)

   # Cloud correction by Berliand (1952)
   strd = strd * (1. + 0.1762 * cc*cc)

   return strd


def lwrad_budyko(lat,tair,e,cc) :
   # downwelling longwave radiation
   #tair : air temperature [K]
   #e    : near surface vapor pressure [Pa]
   #cc   : cloud cover (0-1)

   et=611.*10.**(7.5*(tair-273.16)/(tair-35.86))
#   print numpy.max(tair),numpy.max(et),numpy.max(e)
#   exit(0)
   # Below formula assumes pressure in mBar
   emiss=0.97
   deg2rad = numpy.pi/180.
   # Clear sky downwelling longwave flux from Budyko(1961)
   chi = 0.5+0.246*numpy.abs(lat*deg2rad)
   cc_cliped=numpy.minimum(1.0,numpy.maximum(cc,0.))
   term1=(0.254-4.95e-5*et)*(1.0-chi*cc_cliped**(1.2) )
   strd = emiss*_stefanb * tair**4 * (term1)


   return strd


#MOSTAFA: END


#http://www.nersc.no/biblio/formulation-air-sea-fluxes-esop2-version-micom
#http://regclim.met.no/rapport_4/presentation16/presentation16.htm
#def qlwd(tair,plat,cc,td)
#   fqlw  =emiss*stefanb*tair**3
#   fqlwcc=1.-(.5+.246*abs(plat(i,j)*radian))*cc**1.2
#c
#      fqlwi1=fqlw*tair*((.254-4.95e-5*vpair_i)*fqlwcc-4.)
#      fqlwi2=fqlw*4.
#c
#      fqlww1=fqlw*tair*((.254-4.95e-5*vpair_w)*fqlwcc-4.)
#      fqlww2=fqlwi2

def strd_efimova_jacobs(tair,e,cc) :
   #tair : air temperature [K]
   #e    : near surface vapor pressure [Pa]
   #cc   : cloud cover (0-1)

   # Below formula assumes pressure in mBar
   e_mbar = e * 0.01

   # Clear sky downwelling longwave flux from Eimova(1961)
   strd = _stefanb * tair**4 * (0.746+0.0066*e_mbar) 

   # Cloud correction by Jacobs(1978)
   strd = strd * (1. + 0.26 * cc) 

   return strd

def strd_maykut_jacobs(tair,e,cc) :
   #tair : air temperature [K]
   #cc   : cloud cover (0-1)

   # Below formula assumes pressure in mBar
   e_mbar = e * 0.01

   # Clear sky downwelling longwave flux from Maykut and Church (1973). 
   strd = _stefanb * tair**4 * 0.7855

   # Cloud correction by Jacobs(1978)
   strd = strd * (1. + 0.26 * cc) 

   return strd

def windstress(uwind,vwind) :
   karalight=True
   ws=numpy.sqrt(uwind**2+vwind**2)
   if karalight :
      wndfac=numpy.maximum(2.5,numpy.minimum(32.5,ws))
      cd_new = 1.0E-3*(.692 + .0710*wndfac - .000700*wndfac**2)
   # else :
   #    wndfac=(1.+numpy.sign(1.,ws-11.))*.5
   #    cd_new=(0.49+0.065*ws)*1.0e-3*wndfac+cd*(1.-wndfac)
   wfact=ws*_airdns*cd_new
   taux = uwind*wfact
   tauy = vwind*wfact
   ustar = numpy.sqrt((taux**2+tauy**2)*1e-3)
   return taux, tauy

def vapmix(e,p) :
   # Input is :
   # e = vapour pressure (partial pressure of vapour)
   # p = air pressure
   vapmix = 0.622 * e / (p-e)
   return vapmix

def satvap(t) :
   # This function calculates the saturation vapour pressure
   # [Pa] from the temperature [deg K].
   # Modified: Anita Jacob, June '97
   #
   # Input: t: temperature [deg K]
   # Output: satvap: saturation vapour pressure at temp. t
   #
   # es(T) = C1 * exp(C3*(T - T0)/(T - C4)) from ECMWF manual
   #data c1/610.78/,t00/273.16/
   c1=610.78
   t00=273.16

   #if (t < t00) then
   #   c3 = 21.875
   #   c4 = 7.66
   #else
   #   c3 = 17.269
   #   c4 = 35.86
   #endif
   #KAL !!! c3 = numpy.where(t < t00,21.875,7.66)
   #KAL !!! c4 = numpy.where(t < t00,17.269,35.86)

   # Old hycom
   #c3 = numpy.where(t < t00, 21.875,17.269)
   #c4 = numpy.where(t < t00,  7.66, 35.86)

   # From newest IFS (CY41R2)
   c3 = numpy.where(t < t00, 22.587,17.502)
   c4 = numpy.where(t < t00, -0.7  ,32.19)

   aa = c3 * (t - t00)
   bb = t - c4
   cc=aa/bb

   #if (cc < -20.0) then
   #   satvap=0.0
   #else
   #   satvap = c1 * exp(aa/bb)
   satvap=numpy.where(cc < -20.0, 0.0, c1 * numpy.exp(aa/bb))
   return satvap

def  relhumid(sva,svd,msl) :
   # This routine calculates the relative humidity by the
   # dew point temperature and the mean sea level pressure.
   # Modified: Anita Jacob, June '97
   # Input:
   #    sva: saturatn vapour press at air temp [K]
   #    svd: saturatn vapour press at dew pt temp [K]
   #    msl: pressure at mean sea level [Pa]
   # Output: 
   #   relhumid: Relative Humidity

   # We use the Tetens formula:
   # es(T) = C1 * exp(C3*(T - T0)/(T - C4)) from ECMWF manual
   #              es(Tdew)        p - es(Tair)
   # RH = 100 *  -----------   *  ------------
   #             p - es(tdew)       es(Tair)
   aaa=msl - svd
   aaa = svd/aaa
   bbb = (msl - sva)/sva
   relhumid = 100. * aaa * bbb
   return relhumid

def qsw_allsky_rosato(srad_top,cosz,cosz_noon,cc) :
   # Follows Rosato and Miyakoda[1988]
   # srad = cloud-top incident radiation
   # cosz = cosine of solar zenith angle

   # direct component
   sdir=srad_top*0.7**(1./(cosz+1e-2))     #direct radiation component
   sdif=((1.-_absh2o)*srad_top-sdir)*.5        #diffusive radiation component

   # Solar altitude
   altdeg=numpy.maximum(0.,numpy.arcsin(cosz_noon))*180./numpy.pi #solar noon altitude in degrees

   cfac=(1.-0.62*cc+0.0019*altdeg)               #cloudiness correction by Reed(1977)
   ssurf=(sdir+sdif)*cfac

   return ssurf









def qsw0(qswtime,daysinyear,cc,plat,plon) :
   #
   # --- -------------------------------------------------------------------
   # --- compute 24 hrs mean solar irrradiance at the marine surface layer
   # --- (unit: w/m^2)
   # --- -------------------------------------------------------------------
   #
   # --- Average number of days in year over a 400-year cycle (Gregorian Calendar)
   daysinyear400=365.2425
   #c --- set various quantities
   pi2=8.*numpy.arctan(1.)          #        2 times pi
   deg=360./pi2             #        convert from radians to degrees
   rad=pi2/360.             #        convert from degrees to radians
   eepsil=1.e-9             #        small number
   ifrac=24                 #        split each 12 hrs day into ifrac parts
   fraci=1./ifrac           #        1 over ifrac
   absh2o=0.09              # ---    absorption of water and ozone
   s0=1365.                 # w/m^2  solar constant
   radian=rad
#c
#c --- -------------------------------------------------------------------
#c --- compute 24 hrs mean solar radiation at the marine surface layer 
#c --- -------------------------------------------------------------------
#C --- KAL: TODO - adhere to hycom time setup
   day=numpy.mod(qswtime,daysinyear)    #0 < day < 364
   day=numpy.floor(day)
#c
   dangle=pi2*day/float(daysinyear)   #day-number-angle, in radians 
   if day<0. or day>daysinyear+1 :
      print('qsw0: Error in day for day angle')
      print('Day angle is ',day,daysinyear,qswtime)
      raise NameError("test")
      
# --- compute astronomic quantities -- 
   decli=.006918+.070257*numpy.sin(dangle)   -.399912*numpy.cos(dangle)      \
                +.000907*numpy.sin(2.*dangle)-.006758*numpy.cos(2.*dangle)   \
                +.001480*numpy.sin(3.*dangle)-.002697*numpy.cos(3.*dangle)

   sundv=1.00011+.001280*numpy.sin(dangle)   +.034221*numpy.cos(dangle)      \
                +.000077*numpy.sin(2.*dangle)+.000719*numpy.cos(2.*dangle)

# --- compute astronomic quantities

   sin2=numpy.sin(plat*radian)*numpy.sin(decli)
   cos2=numpy.cos(plat*radian)*numpy.cos(decli)
#
# --- split each day into ifrac parts, and compute the solar radiance for 
# --- each part. by assuming symmetry of the irradiance about noon, it
# --- is sufficient to compute the irradiance for the first 12 hrs of
# --- the (24 hrs) day (mean for the first 12 hrs equals then the mean
# --- for the last 12 hrs)
#
# --- TODO - This routine can also return daily varying solar heat flux
   scosz=0.
   stot=0.
   for npart in range(1,25) :
      bioday=day+(npart-.5)*fraci*.5
      biohr=bioday*86400.                #hour of day in seconds
      biohr=numpy.mod(biohr+43200.,86400.)    #hour of day;  biohr=0  at noon
      hangle=pi2*biohr/86400.            #hour angle, in radians
#
      cosz=numpy.maximum(0.,sin2+cos2*numpy.cos(hangle)) #cosine of the zenith angle
      scosz=scosz+cosz                     #  ..accumulated..
      srad =s0*sundv*cosz                  #extraterrestrial radiation
#
#         sdir=srad*0.7**(1./(cosz+eepsil))    #direct radiation component
#         sdir=srad * exp(-0.356674943938732447/(cosz+eepsil))         
# ---    KAL prevent underflow - .7^100 = 3x10^-16 
      sdir=srad*0.7**(numpy.minimum(100.,1./(cosz+eepsil)))    #direct radiation component
#
      sdif=((1.-absh2o)*srad-sdir)*.5               #diffusive radiation component
      altdeg=numpy.maximum(0.,numpy.arcsin(numpy.minimum(1.0,sin2+cos2)))*deg #solar noon altitude in degrees
      cfac=(1.-0.62*cc+0.0019*altdeg)               #cloudiness correction 
      ssurf=(sdir+sdif)*cfac
      stot=stot+ssurf

#     enddo
   scosz=scosz*fraci               #24-hrs mean of  cosz
   radfl0=stot*fraci               #24-hrs mean shortw rad in w/m^2
#
# --  Original formula was wrong ...
#     !cawdir(i,j)=1.-numpy.maximum(0.15,0.05/(scosz+0.15)) 
#     !cawdir(i,j)=1.-numpy.maximum(0.03,0.05/(scosz+0.15))  !Correction   - Mats
   cawdir=1.-numpy.minimum(0.15,0.05/(scosz+0.15))   #Correction 2 - KAL
#     enddo
#     enddo
#     enddo
#$OMP END PARALLEL DO
#
#     end subroutine qsw0

   return radfl0,cawdir




import importlib
from typing import Type, TYPE_CHECKING

if TYPE_CHECKING:
    from NEDAS.models import Model

registry = {
    'lorenz96': 'Lorenz96Model',
    'qg': 'QGModel',
    # 'qg.emulator': 'QGModelEmulator',
    'vort2d': 'Vort2DModel',
    'topaz.v5': 'Topaz5Model',
    'nextsim.v1': 'NextsimModel',
    'nextsim.dg': 'NextsimDGModel',
    'wrf': 'WRFModel',
}

def get_model_class(model_name: str) -> Type["Model"]:
    """
    Factory function to return the correct Model subclass.

    Args:
        model_name (str): Model name

    Returns:
        Type[Model]: Corresponding Model subclass
    """
    model_name = model_name.lower()

    if model_name not in registry.keys():
        raise NotImplementedError(f"Model class not implemented for '{model_name}'")

    module = importlib.import_module('NEDAS.models.'+model_name)
    ModelClass = getattr(module, registry[model_name])

    return ModelClass



import os
import subprocess
import tempfile
from time import sleep
from NEDAS.utils.conversion import seconds_to_timestr
from NEDAS.utils.progress import find_keyword_in_file, count_lines_in_file
from NEDAS.job_submitters.base import JobSubmitter

class SLURMJobSubmitter(JobSubmitter):
    """JobSubmitter Class customized for SLURM schedulers"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        ##additional slurm options
        self.mem_per_cpu = kwargs.get('mem_per_cpu')

        self.log_file = kwargs.get('log_file', None)
        self.stagnant_log_timeout = kwargs.get('stagnant_log_timeout', 600)

    @property
    def nproc_avail(self):
        return int(os.environ['SLURM_NTASKS'])

    @property
    def nnode_avail(self):
        return int(os.environ['SLURM_NNODES'])

    @property
    def ppn_avail(self):
        return int(os.environ['SLURM_TASKS_PER_NODE'].split('(')[0])

    @property
    def execute_command(self):
        if self.run_separate_jobs:
            return f"srun -n {self.nproc} --unbuffered"
        else:
            if self.parallel_mode == 'mpi':
                return f"srun -n {self.nproc} -N {self.nnode} -r {self.offset_node} --exact --unbuffered"
            elif self.parallel_mode == 'openmp':
                return f"export OMP_NUM_THREADS={self.nproc}; srun -N 1 -r {self.offset_node} -n 1 --cpus-per-task={self.nproc} --unbuffered"
            else:
                raise ValueError(f"unknown parallel_mode '{self.parallel_mode}'")

    @property
    def job_array_index_name(self):
        return '$SLURM_ARRAY_TASK_ID'

    def submit_job_and_monitor(self, commands):
        with tempfile.NamedTemporaryFile(mode='w+', delete=False,
                                         dir=self.run_dir,
                                         prefix=self.job_name+'.',
                                         suffix='.sh') as job_script:
            job_script.write("#!/bin/bash\n")

            ##slurm job header
            job_script.write(f"#SBATCH --job-name={self.job_name}\n")
            job_script.write(f"#SBATCH --account={self.project}\n")
            job_script.write(f"#SBATCH --time={seconds_to_timestr(self.walltime)}\n")
            job_script.write(f"#SBATCH --nodes={self.nnode}\n")
            job_script.write(f"#SBATCH --ntasks-per-node={self.ppn}\n")
            if self.queue and self.queue != 'normal':
                job_script.write(f"#SBATCH --qos={self.queue}\n")
            if self.mem_per_cpu:
                job_script.write(f"#SBATCH --mem-per-cpu={self.mem_per_cpu}\n")

            if self.use_job_array:
                log_file = os.path.join(self.run_dir, f"{self.job_name}-%A_%a.out")
            else:
                log_file = os.path.join(self.run_dir, f"{self.job_name}-%j.out")
            job_script.write(f"#SBATCH --output={log_file}\n")

            if self.use_job_array:
                job_script.write(f"#SBATCH --array=1-{self.array_size}\n")

            ##add the commands
            commands = super().parse_commands(commands)
            job_script.write(commands)
            job_script.write('\n')

            self.job_script = job_script.name
        
        ##submit the job script
        p = subprocess.run(['sbatch', self.job_script], capture_output=True, text=True)
        if p.returncode != 0:
            raise RuntimeError(f"Failed to submit job: {p.stderr}")
        self.job_id = int(p.stdout.split()[-1])

        if self.debug:
            print(f"JobSubmitter: job '{self.job_name}' submitted with ID {self.job_id} to SLURM scheduler", flush=True)

        ##monitor job status
        if self.use_job_array:
            while True:
                sleep(self.check_dt)
                job_finished = []
                for i in range(self.array_size):
                    p = subprocess.run(['squeue', '-h', '-j', f'{self.job_id}_{i}'], capture_output=True, text=True)
                    if not p.stdout:
                        job_finished.append(True)
                    else:
                        job_finished.append(False)    
                if all(job_finished):
                    break
                        
        else:
            elapsed_time = 0
            n0 = 0
            while True:
                sleep(self.check_dt)
                p = subprocess.run(['squeue', '-h', '-j', f'{self.job_id}'], capture_output=True, text=True)
                if not p.stdout:
                    ##job no longer in queue
                    break
                job_status = p.stdout.split()[4]
                if job_status not in ['R', 'PD', 'CG']:
                    ##job not running, pending, or cleaning up
                    raise RuntimeError(f"job {self.job_name} failed with status {job_status}")

                if job_status == 'PD':  ##if job is pending in queue, keep waiting
                    continue

                ##if self.log_file is specified
                ##monitor it, if it becomes stagnant, kill the job and raise error
                if self.log_file is None:
                    continue
                elapsed_time += self.check_dt
                n1 = count_lines_in_file(self.log_file)
                if n1 > n0:
                    elapsed_time = 0
                    n0 = n1
                if elapsed_time > self.stagnant_log_timeout:
                    subprocess.run(['scancel', str(self.job_id)])
                    print(self.job_name, 'stagnant', elapsed_time)
                    raise RuntimeError(f"job {self.job_name} killed: {self.log_file} remain stagnent for {self.stagnant_log_timeout} seconds")

        if self.debug:
            print(f"JobSubmitter: job '{self.job_name}' finished", flush=True)

        ##check log file and report errors
        if self.use_job_array:
            for i in range(self.array_size):
                log_file = os.path.join(self.run_dir, f"{self.job_name}-{self.job_id}_{i}.out")
                if not find_keyword_in_file(log_file, "Job exited normally"):
                    raise RuntimeError(f"job {self.job_name} failed, check {log_file}")
        else:
            log_file = os.path.join(self.run_dir, f"{self.job_name}-{self.job_id}.out")
            if not find_keyword_in_file(log_file, "Job exited normally"):
                raise RuntimeError(f"job {self.job_name} failed, check {log_file}")



import os
from NEDAS.config import Config

def main():
    c = Config(parse_args=True)

    model = c.models['lorenz96']

    ens_init_dir = model.ens_init_dir
    os.system("mkdir -p "+ens_init_dir)

    for m in range(c.nens):
        opts = { 'path': ens_init_dir, 'name': 'state', 'time': c.time, 'member': m, }
        print(f"generating initial condition for member {m+1}, output to {model.filename(**opts)}")

        state = model.generate_initial_condition()

        model.write_var(state, **opts)

    print("done.")

if __name__ == '__main__':
    main()



import numpy as np
from NEDAS.utils.progress import progress_bar
from NEDAS.assim_tools.inflation.base import Inflation

class MultiplicativeInflation(Inflation):
    def adaptive_prior_inflation(self, c, state, obs):
        """compute prior inflate coef by obs-space statistics (Desroziers et al. 2005)"""
        c.print_1p(">>> adaptive prior inflation:\n")
        stats = self.obs_space_stats(c, state, obs)
        if stats['total_nobs'] < 3:
            c.print_1p(f"Warning: insufficient nobs to establish statistics, setting inflate_coef=1")
            self.coef = 1.
        else:
            varb = stats['varb'] / stats['total_nobs']
            varo = stats['varo'] / stats['total_nobs']
            omb2 = stats['omb2'] / stats['total_nobs']
            c.print_1p(f"varb = {varb}, varo={varo}\n")
            c.print_1p(f"omb2 = {omb2}\n")
            self.coef = np.sqrt((omb2 - varo) / varb)

    def adaptive_post_inflation(self, c, state, obs):
        """compute posterior inflate coef by obs-space statistics (Desroziers et al. 2005) """
        c.print_1p(">>> adaptive posterior inflation:\n")
        stats = self.obs_space_stats(c, state, obs)
        if stats['total_nobs'] < 3:
            c.print_1p(f"Warning: insufficient nobs to establish statistics, setting inflate_coef=1")
            self.coef = 1.
        else:
            varb = stats['varb'] / stats['total_nobs']
            vara = stats['vara'] / stats['total_nobs']
            varo = stats['varo'] / stats['total_nobs']
            omb2 = stats['omb2'] / stats['total_nobs']
            omaamb = stats['omaamb'] / stats['total_nobs']
            amb2 = stats['amb2'] / stats['total_nobs']
            c.print_1p(f"varb = {varb}, vara = {vara}, varo={varo}\n")
            c.print_1p(f"omb2 = {omb2}, omaamb = {omaamb}, amb2={amb2}\n")
            # self.coef = np.sqrt(omaamb/vara)
            self.coef = np.sqrt((omb2-varo-amb2)/vara)

    def apply_inflation(self, c, state, flag):
        if flag == 'prior':
            mean_file = state.prior_mean_file
            fields = state.fields_prior
        elif flag == 'posterior':
            mean_file = state.post_mean_file
            fields = state.fields_post
        else:
            raise ValueError(f"Unknown flag {flag}, should be prior or posterior")

        pid_mem_show = [p for p,lst in state.mem_list.items() if len(lst)>0][0]
        pid_rec_show = [p for p,lst in state.rec_list.items() if len(lst)>0][0]
        c.pid_show = pid_rec_show * c.nproc_mem + pid_mem_show

        c.print_1p(f'>>> inflating {flag} ensemble with multiplicative coef={self.coef}\n')

        ##process the fields, each processor goes through its own subset of
        ##mem_id,rec_id simultaneously
        nm = len(state.mem_list[c.pid_mem])
        nr = len(state.rec_list[c.pid_rec])

        for r, rec_id in enumerate(state.rec_list[c.pid_rec]):

            ##read the mean field with rec_id
            fields_mean = state.read_field(mean_file, c.grid.mask, 0, rec_id)
            for m, mem_id in enumerate(state.mem_list[c.pid_mem]):
                if c.debug:
                    print(f"PID {c.pid:4}: inflating mem{mem_id+1:03}", flush=True)
                else:
                    c.print_1p(progress_bar(m*nr+r, nm*nr))

                ##inflate the ensemble perturbations by coef
                fields[mem_id, rec_id] = self.coef*(fields[mem_id, rec_id] - fields_mean) + fields_mean

        c.comm.Barrier()
        c.print_1p(' done.\n')



###adapted from NERSC-HYCOM-CICE/pythonlibs/modelgrid
import numpy as np
from pyproj import Geod

# some constants
_pi_1=np.pi
_pi_2=_pi_1/2.
_deg=180./_pi_1
_rad=1.0/_deg
_epsil=1.0E-9

class ConformalMapping:
    def __init__(self,lat_a,lon_a,lat_b,lon_b,wlim,elim,ires,slim,nlim,jres,mercator,mercfac,lold) :
        """Constructor: arguments: (grid.info)
            lat_a, lon_a      : position of pole A in geo coordinates
            lat_b, lon_b      : position of pole B in geo coordinates
            wlim, elim, ires : western,  eastern  limits in new coords and number of points in 1st dim
            slim, nlim, jres : southern, northern limits in new coords and number of points in 2nd dim
            mercator
            mercfac, lold
        """
        self._lat_a = lat_a
        self._lon_a = lon_a
        self._lat_b = lat_b
        self._lon_b = lon_b

        self._wlim = wlim
        self._elim = elim
        self._ires = ires

        self._slim = slim
        self._nlim = nlim
        self._jres = jres

        self._mercator = mercator
        self._mercfac  = mercfac
        self._lold     = lold

        self._di=(self._elim-self._wlim)/float(ires-1)    # delta lon'
        self._dj=(self._nlim-self._slim)/float(jres-1)    # delta lat' for spherical grid

        if self._mercator:
            self._dj=self._di
            self._dm=self._di
            if lold:
                self._slim=-self._mercfac*self._jres*self._dj
            else:
                self._slim= self._mercfac

        # transform to spherical coordinates
        self._theta_a=self._lon_a*_rad
        self._phi_a=_pi_2-self._lat_a*_rad
        self._theta_b=self._lon_b*_rad
        self._phi_b=_pi_2-self._lat_b*_rad

        # find the angles of a vector pointing at a point located exactly
        # between the poles
        cx=np.cos(self._theta_a)*np.sin(self._phi_a)+np.cos(self._theta_b)*np.sin(self._phi_b)
        cy=np.sin(self._theta_a)*np.sin(self._phi_a)+np.sin(self._theta_b)*np.sin(self._phi_b)
        cz=np.cos(self._phi_a)+np.cos(self._phi_b)

        theta_c=np.arctan2(cy,cx)
        self._phi_c=_pi_2-np.arctan2(cz,np.sqrt(cx*cx+cy*cy))

        # initialize constants used in the conformal mapping
        self._imag=complex(.0,1.)
        self._ac=np.tan(.5*self._phi_a)*np.exp(self._imag*self._theta_a)
        self._bc=np.tan(.5*self._phi_b)*np.exp(self._imag*self._theta_b)
        self._c =np.tan(.5*self._phi_c)*np.exp(self._imag*theta_c)
        self._cmna=self._c-self._ac
        self._cmnb=self._c-self._bc

        w=self._cmnb/self._cmna
        self._mu_s=np.arctan2(np.imag(w),np.real(w))
        self._psi_s=2.*np.arctan(abs(w))

        ##grid spacing in meters
        ii, jj = np.meshgrid(np.arange(self._ires), np.arange(self._jres))
        lat, lon = self.gind2ll(ii+1., jj+1.)
        geod = Geod(ellps='sphere')
        _, _, dist_x = geod.inv(lon[:, 1:], lat[:, 1:], lon[:, :-1], lat[:, :-1])
        _, _, dist_y = geod.inv(lon[1:, :], lat[1:, :], lon[:-1, :], lat[:-1, :])
        self._dx = np.median(dist_x)
        self._dy = np.median(dist_y)

    def __call__(self, x, y, inverse=False):
        """
        mimick the pyproj.Proj object, when called proj can perform mapping from long/lat to x/y and inverse)
        """
        if not inverse:
            i, j = self.ll2gind(y, x)
            xo = (i-1) * self._dx
            yo = (j-1) * self._dy
        else:
            i = np.atleast_1d(x / self._dx + 1)
            j = np.atleast_1d(y / self._dy + 1)
            yo, xo = self.gind2ll(i, j)
        if xo.size == 1:
            return xo.item(), yo.item()
        return xo, yo

    def __eq__(self, other):
        """
        Define when two confmapping object is the same
        """
        if not isinstance(other, ConformalMapping):
            return False

        for attr in ['lat_a', 'lon_a', 'lat_b', 'lon_b', 'wlim', 'elim', 'ires', 'slim', 'nlim', 'jres', 'mercator', 'mercfac', 'lold']:
            if getattr(self, '_'+attr) != getattr(other, '_'+attr):
                return False
        return True

    @classmethod
    def init_from_file(cls,filename) :
        fid=open(filename,"r")

        #-40.0 140.0      ! Position of pole N (lattitude, longitude):
        tmp=fid.readline().split("!")[0].strip().split()
        lat_a,lon_a = [float(elem) for elem in tmp ]

        #-50.0 140.0        ! Position of pole S (lattitude, longitude):
        tmp=fid.readline().split("!")[0].strip().split()
        lat_b,lon_b = [float(elem) for elem in tmp ]

        #178.2 181.42 800 ! Longitude interval West lon,    East lon, idim
        tmp=fid.readline().split("!")[0].strip().split()
        wlim,elim = [float(elem) for elem in tmp[0:2]]
        ires        = int(tmp[2])

        # 0.0  80.0 760 ! Lattitude interval south limit, north limit, jdim
        tmp=fid.readline().split("!")[0].strip().split()
        slim,nlim = [float(elem) for elem in tmp[0:2]]
        jres        = int(tmp[2])

        #.true.                ! Generate topography
        fid.readline() # Not needed for grid generation

        #.true.                ! dump teclatlon.dat
        fid.readline() # Not needed for grid generation

        #.true.                ! dump micom latlon.dat
        fid.readline() # Not needed for grid generation

        #.true.                ! mercator grid (true, false)
        tmp=fid.readline().split("!")[0].strip().split()
        if tmp[0] == ".true." :
            mercator=True
        elif tmp[0] == ".false." :
            mercator=False
        else :
            raise ValueError("Unable to safely resolve value of mercator flag for confmap")

        #    0.365 .false.    ! merc fac
        tmp=fid.readline().split("!")[0].strip().split()
        mercfac = float(tmp[0])
        if tmp[1] == ".true." :
            lold=True
        elif tmp[1] == ".false." :
            lold=False
        else :
            raise ValueError("Unable to safely resolve value of lold flag for confmap")

        #.false.              ! Smoothing, Shapiro filter
        fid.readline() # Not needed for grid generation

        # 8    2                ! Order of Shapiro filter,  number of passes
        fid.readline() # Not needed for grid generation

        fid.close()

        return cls(lat_a,lon_a,lat_b,lon_b,
                   wlim,elim,ires,
                   slim,nlim,jres,
                   mercator,
                   mercfac,lold)

    def oldtonew(self,lat_o,lon_o) :
        # this routine performes a conformal mapping of the old to the new
        # coordinate system
        lat_o=np.atleast_1d(lat_o)
        lon_o=np.atleast_1d(lon_o)

        # transform to spherical coordinates
        theta=np.mod(lon_o*_rad+3.0*_pi_1,2.0*_pi_1)-_pi_1
        phi=_pi_2-lat_o*_rad

        # transform to the new coordinate system: 1)
        z=np.tan(.5*phi)*np.exp(self._imag*theta)
        w=(z-self._ac)*self._cmnb/((z-self._bc)*self._cmna)
        mu=np.arctan2(np.imag(w),np.real(w))
        psi=2.*np.arctan(abs(w))

        # transform to the new coordinate system: 2)
        I=np.abs(phi-_pi_1)<_epsil
        mu[I]=self._mu_s
        psi[I]=self._psi_s

        # transform to the new coordinate system: 3)
        I=np.logical_and(np.abs(phi-self._phi_b)<_epsil, np.abs(theta-self._theta_b)<_epsil)
        mu[I]=0.
        psi[I]=_pi_1

        # transform to new lat/lon coordinates
        lat_n=(_pi_2-psi)*_deg
        lon_n=mu*_deg

        return lat_n,lon_n

    def newtoold(self,lat_n,lon_n) :
        # this routine performes a conformal mapping of the new to the old 
        # coordinate system

        # transform to spherical coordinates
        mu=np.mod(lon_n*_rad+3*_pi_1,2*_pi_1)-_pi_1
        psi=np.abs(_pi_2-lat_n*_rad)

        # transform to the old coordinate system
        w=np.tan(.5*psi)*np.exp(self._imag*mu)
        z=(self._ac*self._cmnb-w*self._bc*self._cmna)/(self._cmnb-w*self._cmna)
        theta=np.arctan2(np.imag(z),np.real(z))
        phi=2.*np.arctan(np.abs(z))

        I = np.abs(psi-_pi_1) < _epsil
        theta[I]=self._theta_b
        phi  [I]=self._phi_b

        I = np.logical_and(
            np.abs(mu-self._mu_s)<_epsil,
            (psi-self._psi_s)<_epsil)
        theta[I]=0.
        phi  [I]=_pi_1

        # transform to lat/lon coordinates
        lat_o=(_pi_2-phi)*_deg
        lon_o=theta*_deg

        return lat_o,lon_o

    def pivotp(self,lat_n,lon_n) :
        # This subroutine computes the pivot point of each of the observations
        # in the temporary array tmpobs of type observation. The pivot point
        # is the biggest i and the biggest j, (i,j) is the computation points/
        # the grid, that is less than the position of the observation.
        lat_n=np.atleast_1d(lat_n)
        lon_n=np.atleast_1d(lon_n)

        # fix for wrap-around:
        lon_n[np.where(lon_n<0)] += 360.

        ipiv=(lon_n-self._wlim)/self._di + 1.0

        jpiv=np.ones(ipiv.shape)*-999
        if self._mercator:
            I=np.abs(lat_n) < 89.999
            tmptan=np.tan(0.5*_rad*lat_n[I]+0.25*_pi_1)
            jpiv[I]=(np.log(tmptan)-self._slim*_rad)/(_rad*self._dj) + 1.0
        else :
            jpiv=(lat_n-self._slim)/self._dj + 1.0

        # Returns floats, cast to int to use as index/pivot points
        return ipiv,jpiv

    def get_grid_point(self,i,j,shifti=0.,shiftj=0.) :
        #! Used when creating a grid
        #! Retrieves  lon/lat for grid index (i,j) at P points
        #lon_n=self._wlim+(i-1)self._di+shifti
        #lat_n=self._slim+(j-1)*self._dj+shiftj
        lon_n=self._wlim+(i-1+shifti)*self._di
        lat_n=self._slim+(j-1+shiftj)*self._dj
        if self._mercator :
            lat_n=self._slim+(j-1+shiftj)*self._dm
            lat_n=(2.*np.arctan(np.exp((lat_n*_rad)))-_pi_1*.5)*_deg
        return self.newtoold(lat_n,lon_n)

    def ll2gind(self,lat_o,lon_o) :
        """ Returns grid index (floats) of specified lat and lon coordinates on the grid"""
        lat_n,lon_n = self.oldtonew(lat_o,lon_o)
        return self.pivotp(lat_n,lon_n)

    def gind2ll(self,i,j,shifti=0.,shiftj=0.):
        """ Returns lat and lon for grid index i,j """
        return self.get_grid_point(i,j,shifti=shifti,shiftj=shiftj)

